__all__ = ["BybitClient"]

import json
import time
from typing import Any

from unicex._base import BaseClient
from unicex.exceptions import NotAuthorized
from unicex.types import RequestMethod
from unicex.utils import filter_params, generate_hmac_sha256_signature

from .types import Category, FuturesCategory, OrderType, Side, Timeframe


class _BaseBybitClient(BaseClient):
    """Базовый класс для клиентов Bybit API."""

    _BASE_URL: str = "https://api.bybit.com"
    """Базовый URL для REST API Bybit."""

    _RECV_WINDOW: str = "5000"
    """Стандартный интервал времени для получения ответа от сервера."""

    def _get_headers(self, timestamp: str, signature: str | None = None) -> dict:
        """Возвращает заголовки для запросов к Bybit API.

        Параметры:
            timestamp (str): Временная метка запроса в миллисекундах.
            signature (str | None): Подпись запроса, если запрос авторизированый.
        """
        headers = {"Accept": "application/json", "Content-Type": "application/json"}
        if signature:
            headers["X-BAPI-API-KEY"] = self._api_key  # type: ignore
            headers["X-BAPI-SIGN-TYPE"] = "2"
            headers["X-BAPI-SIGN"] = signature
            headers["X-BAPI-RECV-WINDOW"] = self._RECV_WINDOW
            headers["X-BAPI-TIMESTAMP"] = timestamp
        return headers

    def _make_request(
        self,
        method: RequestMethod,
        url: str,
        signed: bool = False,
        *,
        params: dict[str, Any] | None = None,
    ) -> Any:
        """Выполняет HTTP-запрос к эндпоинтам Bybit API с поддержкой подписи.

        Если signed=True, формируется подпись для приватных endpoint'ов.
        Если signed=False, запрос отправляется как обычный публичный, через
        базовый _make_request без обработки подписи.

        Параметры:
            method (str): HTTP метод запроса ("GET", "POST", "DELETE" и т.д.).
            url (str): Полный URL эндпоинта Bybit API.
            signed (bool): Нужно ли подписывать запрос.
            params (dict | None): Параметры запроса. Передаются в body, если запрос типа "POST", иначе в query_params

        Возвращает:
            dict: Ответ в формате JSON.
        """
        # Фильтруем параметры от None значений
        params = filter_params(params) if params else {}

        # Генерируем временную метку
        timestamp = str(int(time.time() * 1000))

        # Проверяем нужно ли подписывать запрос
        if not signed:
            headers = self._get_headers(timestamp)
            return super()._make_request(
                method=method,
                url=url,
                headers=headers,
                params=params,
            )

        # Проверяем наличие апи ключей для подписи запроса
        if not self._api_key or not self._api_secret:
            raise NotAuthorized("Api key is required to private endpoints")

        # Формируем payload
        payload = params

        # Генерируем строку для подписи
        # Источник: https://github.com/bybit-exchange/api-usage-examples/blob/master/V5_demo/api_demo/Encryption_HMAC.py
        dumped_payload = json.dumps(payload)
        prepared_query_string = timestamp + self._api_key + self._RECV_WINDOW + dumped_payload
        signature = generate_hmac_sha256_signature(self._api_secret, prepared_query_string)

        # Генерируем заголовки (вкл. в себя подпись и апи ключ)
        headers = self._get_headers(timestamp, signature)

        if method == "POST":  # Отправляем параметры в тело запроса
            return super()._make_request(
                method=method,
                url=url,
                json=payload,
                headers=headers,
            )
        else:  # Иначе параметры добавляем к query string
            return super()._make_request(
                method=method,
                url=url,
                params=payload,
                headers=headers,
            )


class BybitClient(_BaseBybitClient):
    """Клиент для работы с Bybit API (REST v5)."""

    # ==========================================================================
    # MARKET DATA (Public)
    # ==========================================================================

    def ping(self) -> dict:
        """Проверка подключения к REST API.

        https://bybit-exchange.github.io/docs/v5/market/time
        """
        url = self._BASE_URL + "/v5/market/time"
        return self._make_request("GET", url)

    def instruments_info(
        self,
        category: Category,
        symbol: str | None = None,
        status: str | None = None,
        base_coin: str | None = None,
        limit: int | None = None,
        cursor: str | None = None,
    ) -> dict:
        """Информация об инструментах торговли.

        https://bybit-exchange.github.io/docs/v5/market/instrument
        """
        url = self._BASE_URL + "/v5/market/instruments-info"
        params = {
            "category": category,
            "symbol": symbol,
            "status": status,
            "baseCoin": base_coin,
            "limit": limit,
            "cursor": cursor,
        }

        return self._make_request("GET", url, params=params)

    def tickers(
        self,
        category: Category,
        symbol: str | None = None,
        base_coin: str | None = None,
        exp_date: str | None = None,
    ) -> dict:
        """Тикеры (вкл. 24h stats и last price).

        https://bybit-exchange.github.io/docs/v5/market/tickers
        """
        url = self._BASE_URL + "/v5/market/tickers"
        params = {
            "category": category,
            "symbol": symbol,
            "baseCoin": base_coin,
            "expDate": exp_date,
        }

        return self._make_request("GET", url, params=params)

    def mark_price_kline(
        self,
        category: Category,
        symbol: str,
        interval: Timeframe,
        start: int | None = None,
        end: int | None = None,
        limit: int | None = None,
    ) -> dict:
        """Свечи по mark price.

        https://bybit-exchange.github.io/docs/v5/market/mark-kline
        """
        url = self._BASE_URL + "/v5/market/mark-price-kline"
        params = {
            "category": category,
            "symbol": symbol,
            "interval": interval,
            "start": start,
            "end": end,
            "limit": limit,
        }

        return self._make_request("GET", url, params=params)

    def index_price_kline(
        self,
        category: Category,
        symbol: str,
        interval: Timeframe,
        start: int | None = None,
        end: int | None = None,
        limit: int | None = None,
    ) -> dict:
        """Свечи по index price.

        https://bybit-exchange.github.io/docs/v5/market/index-kline
        """
        url = self._BASE_URL + "/v5/market/index-price-kline"
        params = {
            "category": category,
            "symbol": symbol,
            "interval": interval,
            "start": start,
            "end": end,
            "limit": limit,
        }

        return self._make_request("GET", url, params=params)

    def premium_index_price_kline(
        self,
        category: Category,
        symbol: str,
        interval: Timeframe,
        start: int | None = None,
        end: int | None = None,
        limit: int | None = None,
    ) -> dict:
        """Свечи по premium index price.

        https://bybit-exchange.github.io/docs/v5/market/premium-index-kline
        """
        url = self._BASE_URL + "/v5/market/premium-index-price-kline"
        params = {
            "category": category,
            "symbol": symbol,
            "interval": interval,
            "start": start,
            "end": end,
            "limit": limit,
        }

        return self._make_request("GET", url, params=params)

    def klines(
        self,
        symbol: str,
        interval: Timeframe,
        category: Category,
        start: int | None = None,
        end: int | None = None,
        limit: int | None = None,
    ) -> dict:
        """Исторические свечи.

        https://bybit-exchange.github.io/docs/v5/market/kline
        """
        url = self._BASE_URL + "/v5/market/kline"
        params = {
            "category": category,
            "symbol": symbol,
            "interval": interval,
            "start": start,
            "end": end,
            "limit": limit,
        }

        return self._make_request("GET", url, params=params)

    def recent_trades(
        self,
        category: Category,
        symbol: str,
        limit: int | None = None,
    ) -> dict:
        """Последние сделки.

        https://bybit-exchange.github.io/docs/v5/market/recent-trade
        """
        url = self._BASE_URL + "/v5/market/recent-trade"
        params = {"category": category, "symbol": symbol, "limit": limit}

        return self._make_request("GET", url, params=params)

    def orderbook(
        self,
        category: Category,
        symbol: str,
        limit: int | None = None,
    ) -> dict:
        """Стакан ордеров (Order Book).

        https://bybit-exchange.github.io/docs/v5/market/orderbook
        """
        url = self._BASE_URL + "/v5/market/orderbook"
        params = {"category": category, "symbol": symbol, "limit": limit}

        return self._make_request("GET", url, params=params)

    def funding_history(
        self,
        category: Category,
        symbol: str | None = None,
        start_time: int | None = None,
        end_time: int | None = None,
        limit: int | None = None,
    ) -> dict:
        """История фондирования.

        https://bybit-exchange.github.io/docs/v5/market/funding
        """
        url = self._BASE_URL + "/v5/market/funding/history"
        params = {
            "category": category,
            "symbol": symbol,
            "startTime": start_time,
            "endTime": end_time,
            "limit": limit,
        }

        return self._make_request("GET", url, params=params)

    def open_interest(
        self,
        category: Category,
        symbol: str | None = None,
        interval_time: str | None = None,
        start_time: int | None = None,
        end_time: int | None = None,
        limit: int | None = None,
    ) -> dict:
        """Открытый интерес.

        https://bybit-exchange.github.io/docs/v5/market/open-interest
        """
        url = self._BASE_URL + "/v5/market/open-interest"
        params = {
            "category": category,
            "symbol": symbol,
            "intervalTime": interval_time,
            "startTime": start_time,
            "endTime": end_time,
            "limit": limit,
        }

        return self._make_request("GET", url, params=params)

    def insurance(self, coin: str | None = None) -> dict:
        """Данные страхового фонда.

        https://bybit-exchange.github.io/docs/v5/market/insurance
        """
        url = self._BASE_URL + "/v5/market/insurance"
        params = {"coin": coin}

        return self._make_request("GET", url, params=params)

    def risk_limit(self, category: Category, symbol: str | None = None) -> dict:
        """Лимиты риска по инструментам.

        https://bybit-exchange.github.io/docs/v5/market/risk-limit
        """
        url = self._BASE_URL + "/v5/market/risk-limit"
        params = {"category": category, "symbol": symbol}

        return self._make_request("GET", url, params=params)

    def delivery_price(self, category: Category, symbol: str) -> dict:
        """Delivery price (для опционов/фьючерсов).

        https://bybit-exchange.github.io/docs/v5/market/delivery-price
        """
        url = self._BASE_URL + "/v5/market/delivery-price"
        params = {"category": category, "symbol": symbol}

        return self._make_request("GET", url, params=params)

    # ==========================================================================
    # ORDER (Private)
    # ==========================================================================

    def create_order(
        self,
        category: Category,
        symbol: str,
        side: Side,
        orderType: OrderType,
        qty: str,
        price: str | None = None,
        timeInForce: str | None = None,
        orderLinkId: str | None = None,
        isLeverage: int | None = None,
        orderFilter: str | None = None,
        triggerPrice: str | None = None,
        triggerDirection: int | None = None,
        marketUnit: str | None = None,
        slippageToleranceType: str | None = None,
        slippageTolerance: str | None = None,
        triggerBy: str | None = None,
        orderIv: str | None = None,
        positionIdx: int | None = None,
        takeProfit: str | None = None,
        stopLoss: str | None = None,
        tpTriggerBy: str | None = None,
        slTriggerBy: str | None = None,
        tpslMode: str | None = None,
        tpLimitPrice: str | None = None,
        slLimitPrice: str | None = None,
        tpOrderType: str | None = None,
        slOrderType: str | None = None,
        reduceOnly: bool | None = None,
        closeOnTrigger: bool | None = None,
        smpType: str | None = None,
        mmp: bool | None = None,
    ) -> dict:
        """Создание ордера.

        https://bybit-exchange.github.io/docs/v5/order/create-order
        """
        url = self._BASE_URL + "/v5/order/create"
        params = {
            "category": category,
            "symbol": symbol,
            "side": side,
            "orderType": orderType,
            "qty": qty,
            "price": price,
            "timeInForce": timeInForce,
            "orderLinkId": orderLinkId,
            "isLeverage": isLeverage,
            "orderFilter": orderFilter,
            "triggerPrice": triggerPrice,
            "triggerDirection": triggerDirection,
            "marketUnit": marketUnit,
            "slippageToleranceType": slippageToleranceType,
            "slippageTolerance": slippageTolerance,
            "triggerBy": triggerBy,
            "orderIv": orderIv,
            "positionIdx": positionIdx,
            "takeProfit": takeProfit,
            "stopLoss": stopLoss,
            "tpTriggerBy": tpTriggerBy,
            "slTriggerBy": slTriggerBy,
            "tpslMode": tpslMode,
            "tpLimitPrice": tpLimitPrice,
            "slLimitPrice": slLimitPrice,
            "tpOrderType": tpOrderType,
            "slOrderType": slOrderType,
            "reduceOnly": reduceOnly,
            "closeOnTrigger": closeOnTrigger,
            "smpType": smpType,
            "mmp": mmp,
        }

        return self._make_request("POST", url, True, params=params)

    def amend_order(
        self,
        category: Category,
        symbol: str,
        orderId: str | None = None,
        orderLinkId: str | None = None,
        qty: str | None = None,
        price: str | None = None,
        triggerPrice: str | None = None,
        takeProfit: str | None = None,
        stopLoss: str | None = None,
        tpTriggerBy: str | None = None,
        slTriggerBy: str | None = None,
        triggerBy: str | None = None,
        tpslMode: str | None = None,
        tpLimitPrice: str | None = None,
        slLimitPrice: str | None = None,
        tpOrderType: str | None = None,
        slOrderType: str | None = None,
        orderIv: str | None = None,
        reduceOnly: bool | None = None,
        closeOnTrigger: bool | None = None,
        smpType: str | None = None,
        mmp: bool | None = None,
    ) -> dict:
        """Изменение ордера.

        https://bybit-exchange.github.io/docs/v5/order/amend-order
        """
        url = self._BASE_URL + "/v5/order/amend"
        params = {
            "category": category,
            "symbol": symbol,
            "orderId": orderId,
            "orderLinkId": orderLinkId,
            "qty": qty,
            "price": price,
            "triggerPrice": triggerPrice,
            "takeProfit": takeProfit,
            "stopLoss": stopLoss,
            "tpTriggerBy": tpTriggerBy,
            "slTriggerBy": slTriggerBy,
            "triggerBy": triggerBy,
            "tpslMode": tpslMode,
            "tpLimitPrice": tpLimitPrice,
            "slLimitPrice": slLimitPrice,
            "tpOrderType": tpOrderType,
            "slOrderType": slOrderType,
            "orderIv": orderIv,
            "reduceOnly": reduceOnly,
            "closeOnTrigger": closeOnTrigger,
            "smpType": smpType,
            "mmp": mmp,
        }

        return self._make_request("POST", url, True, params=params)

    def cancel_order(
        self,
        category: Category,
        symbol: str,
        orderId: str | None = None,
        orderLinkId: str | None = None,
        orderFilter: str | None = None,
    ) -> dict:
        """Отмена ордера.

        https://bybit-exchange.github.io/docs/v5/order/cancel-order
        """
        url = self._BASE_URL + "/v5/order/cancel"
        params = {
            "category": category,
            "symbol": symbol,
            "orderId": orderId,
            "orderLinkId": orderLinkId,
            "orderFilter": orderFilter,
        }

        return self._make_request("POST", url, True, params=params)

    def cancel_all_orders(
        self,
        category: Category,
        symbol: str | None = None,
        base_coin: str | None = None,
        settle_coin: str | None = None,
        orderFilter: str | None = None,
        stopOrderType: str | None = None,
    ) -> dict:
        """Отмена всех ордеров.

        https://bybit-exchange.github.io/docs/v5/order/cancel-all
        """
        url = self._BASE_URL + "/v5/order/cancel-all"
        params = {
            "category": category,
            "symbol": symbol,
            "baseCoin": base_coin,
            "settleCoin": settle_coin,
            "orderFilter": orderFilter,
            "stopOrderType": stopOrderType,
        }

        return self._make_request("POST", url, True, params=params)

    def open_orders(
        self,
        category: Category,
        symbol: str | None = None,
        base_coin: str | None = None,
        settle_coin: str | None = None,
        orderId: str | None = None,
        orderLinkId: str | None = None,
        orderFilter: str | None = None,
        limit: int | None = None,
        cursor: str | None = None,
    ) -> dict:
        """Список открытых ордеров.

        https://bybit-exchange.github.io/docs/v5/order/open-order
        """
        url = self._BASE_URL + "/v5/order/realtime"
        params = {
            "category": category,
            "symbol": symbol,
            "baseCoin": base_coin,
            "settleCoin": settle_coin,
            "orderId": orderId,
            "orderLinkId": orderLinkId,
            "orderFilter": orderFilter,
            "limit": limit,
            "cursor": cursor,
        }

        return self._make_request("GET", url, True, params=params)

    def order_history(
        self,
        category: Category,
        symbol: str | None = None,
        base_coin: str | None = None,
        settle_coin: str | None = None,
        order_id: str | None = None,
        order_link_id: str | None = None,
        limit: int | None = None,
        cursor: str | None = None,
    ) -> dict:
        """История ордеров.

        https://bybit-exchange.github.io/docs/v5/order/order-list
        """
        url = self._BASE_URL + "/v5/order/history"
        params = {
            "category": category,
            "symbol": symbol,
            "baseCoin": base_coin,
            "settleCoin": settle_coin,
            "orderId": order_id,
            "orderLinkId": order_link_id,
            "limit": limit,
            "cursor": cursor,
        }

        return self._make_request("GET", url, signed=True, params=params)

    def batch_create_orders(self, category: Category, request: list[dict]) -> dict:
        """Пакетное создание ордеров.

        https://bybit-exchange.github.io/docs/v5/order/batch-place
        """
        url = self._BASE_URL + "/v5/order/batch-create"
        params = {"category": category, "request": request}

        return self._make_request("POST", url, True, params=params)

    def batch_amend_orders(self, category: Category, request: list[dict]) -> dict:
        """Пакетное изменение ордеров.

        https://bybit-exchange.github.io/docs/v5/order/batch-amend
        """
        url = self._BASE_URL + "/v5/order/batch-amend"
        params = {"category": category, "request": request}

        return self._make_request("POST", url, True, params=params)

    def batch_cancel_orders(self, category: Category, request: list[dict]) -> dict:
        """Пакетная отмена ордеров.

        https://bybit-exchange.github.io/docs/v5/order/batch-cancel
        """
        url = self._BASE_URL + "/v5/order/batch-cancel"
        params = {"category": category, "request": request}

        return self._make_request("POST", url, True, params=params)

    # ==========================================================================
    # POSITION (Private)
    # ==========================================================================

    def positions(
        self,
        category: Category,
        symbol: str | None = None,
        base_coin: str | None = None,
        settle_coin: str | None = None,
        limit: int | None = None,
        cursor: str | None = None,
    ) -> dict:
        """Список позиций.

        https://bybit-exchange.github.io/docs/v5/position/list
        """
        url = self._BASE_URL + "/v5/position/list"
        params = {
            "category": category,
            "symbol": symbol,
            "baseCoin": base_coin,
            "settleCoin": settle_coin,
            "limit": limit,
            "cursor": cursor,
        }

        return self._make_request("GET", url, True, params=params)

    def set_leverage(
        self,
        category: FuturesCategory,
        symbol: str,
        buy_leverage: str,
        sell_leverage: str,
    ) -> dict:
        """Установить кредитное плечо.

        https://bybit-exchange.github.io/docs/v5/position/leverage
        """
        url = self._BASE_URL + "/v5/position/set-leverage"
        params = {
            "category": category,
            "symbol": symbol,
            "buyLeverage": buy_leverage,
            "sellLeverage": sell_leverage,
        }

        return self._make_request("POST", url, True, params=params)

    def switch_isolated(
        self,
        category: FuturesCategory,
        symbol: str,
        tradeMode: int,
        buyLeverage: str | None = None,
        sellLeverage: str | None = None,
    ) -> dict:
        """Переключение режима маржи (cross/isolated).

        https://bybit-exchange.github.io/docs/v5/position/switch-isolated
        """
        url = self._BASE_URL + "/v5/position/switch-isolated"
        params = {
            "category": category,
            "symbol": symbol,
            "tradeMode": tradeMode,
            "buyLeverage": buyLeverage,
            "sellLeverage": sellLeverage,
        }

        return self._make_request("POST", url, True, params=params)

    def set_tpsl_mode(
        self,
        category: FuturesCategory,
        symbol: str,
        tpSlMode: str,
    ) -> dict:
        """Режим TP/SL (partial/full).

        https://bybit-exchange.github.io/docs/v5/position/tpsl-mode
        """
        url = self._BASE_URL + "/v5/position/set-tpsl-mode"
        params = {"category": category, "symbol": symbol, "tpSlMode": tpSlMode}

        return self._make_request("POST", url, True, params=params)

    def set_trading_stop(
        self,
        category: FuturesCategory,
        symbol: str,
        takeProfit: str | None = None,
        stopLoss: str | None = None,
        trailingStop: str | None = None,
        tpTriggerBy: str | None = None,
        slTriggerBy: str | None = None,
        positionIdx: int | None = None,
        tpslMode: str | None = None,
        tpLimitPrice: str | None = None,
        slLimitPrice: str | None = None,
        tpOrderType: str | None = None,
        slOrderType: str | None = None,
    ) -> dict:
        """Установка TP/SL/Trailing Stop.

        https://bybit-exchange.github.io/docs/v5/position/trading-stop
        """
        url = self._BASE_URL + "/v5/position/trading-stop"
        params = {
            "category": category,
            "symbol": symbol,
            "takeProfit": takeProfit,
            "stopLoss": stopLoss,
            "trailingStop": trailingStop,
            "tpTriggerBy": tpTriggerBy,
            "slTriggerBy": slTriggerBy,
            "positionIdx": positionIdx,
            "tpslMode": tpslMode,
            "tpLimitPrice": tpLimitPrice,
            "slLimitPrice": slLimitPrice,
            "tpOrderType": tpOrderType,
            "slOrderType": slOrderType,
        }

        return self._make_request("POST", url, True, params=params)

    def set_risk_limit(
        self,
        category: FuturesCategory,
        symbol: str,
        riskId: int,
        positionIdx: int | None = None,
    ) -> dict:
        """Установка лимита риска.

        https://bybit-exchange.github.io/docs/v5/position/set-risk-limit
        """
        url = self._BASE_URL + "/v5/position/set-risk-limit"
        params = {
            "category": category,
            "symbol": symbol,
            "riskId": riskId,
            "positionIdx": positionIdx,
        }

        return self._make_request("POST", url, True, params=params)

    def set_auto_add_margin(
        self,
        category: FuturesCategory,
        symbol: str,
        autoAddMargin: int,
        positionIdx: int | None = None,
    ) -> dict:
        """Auto add margin (вкл/выкл).

        https://bybit-exchange.github.io/docs/v5/position/auto-add-margin
        """
        url = self._BASE_URL + "/v5/position/set-auto-add-margin"
        params = {
            "category": category,
            "symbol": symbol,
            "autoAddMargin": autoAddMargin,
            "positionIdx": positionIdx,
        }

        return self._make_request("POST", url, True, params=params)

    def switch_position_mode(self, category: FuturesCategory, symbol: str, mode: int) -> dict:
        """Hedge (двунаправленный) / One-way режим.

        https://bybit-exchange.github.io/docs/v5/position/switch-mode
        """
        url = self._BASE_URL + "/v5/position/switch-mode"
        params = {"category": category, "symbol": symbol, "mode": mode}

        return self._make_request("POST", url, True, params=params)

    # ==========================================================================
    # ACCOUNT (Private)
    # ==========================================================================

    def wallet_balance(
        self,
        accountType: str,
        coin: str | None = None,
    ) -> dict:
        """Баланс кошелька.

        https://bybit-exchange.github.io/docs/v5/account/wallet-balance
        """
        url = self._BASE_URL + "/v5/account/wallet-balance"
        params = {"accountType": accountType, "coin": coin}

        return self._make_request("GET", url, True, params=params)

    def fee_rate(
        self, category: Category, symbol: str | None = None, base_coin: str | None = None
    ) -> dict:
        """Торговые комиссии (ваши персональные).

        https://bybit-exchange.github.io/docs/v5/account/fee-rate
        """
        url = self._BASE_URL + "/v5/account/fee-rate"
        params = {"category": category, "symbol": symbol, "baseCoin": base_coin}

        return self._make_request("GET", url, True, params=params)

    def transaction_log(
        self,
        accountType: str,
        category: str | None = None,
        currency: str | None = None,
        base_coin: str | None = None,
        start_time: int | None = None,
        end_time: int | None = None,
        limit: int | None = None,
        cursor: str | None = None,
    ) -> dict:
        """Транзакционный лог (сводка движений средств).

        https://bybit-exchange.github.io/docs/v5/account/transaction-log
        """
        url = self._BASE_URL + "/v5/account/transaction-log"
        params = {
            "accountType": accountType,
            "category": category,
            "currency": currency,
            "baseCoin": base_coin,
            "startTime": start_time,
            "endTime": end_time,
            "limit": limit,
            "cursor": cursor,
        }

        return self._make_request("GET", url, True, params=params)

    def set_margin_mode(
        self,
        setMarginMode: int,
        productType: str,
    ) -> dict:
        """Unified Margin / Portfolio Margin переключение режима (если доступно).

        https://bybit-exchange.github.io/docs/v5/account/set-margin-mode
        """
        url = self._BASE_URL + "/v5/account/set-margin-mode"
        params = {"setMarginMode": setMarginMode, "productType": productType}

        return self._make_request("POST", url, True, params=params)

    def set_fee_rate(
        self,
        symbol: str,
        takerFeeRate: str | None = None,
        makerFeeRate: str | None = None,
    ) -> dict:
        """(Для брокеров/привилегий) Настройка своей комиссии — если доступно.

        https://bybit-exchange.github.io/docs/v5/account/fee-rate
        """
        url = self._BASE_URL + "/v5/account/set-fee-rate"
        params = {
            "symbol": symbol,
            "takerFeeRate": takerFeeRate,
            "makerFeeRate": makerFeeRate,
        }

        return self._make_request("POST", url, True, params=params)

    # ==========================================================================
    # ASSET / WALLET (Private)
    # ==========================================================================

    def deposit_withdraw_status(
        self,
        coin: str | None = None,
    ) -> dict:
        """Статус депозита/вывода по монетам.

        https://bybit-exchange.github.io/docs/v5/asset/coin-info
        """
        url = self._BASE_URL + "/v5/asset/coin/query-info"
        params = {"coin": coin}

        return self._make_request("GET", url, True, params=params)

    def deposit_list(
        self,
        coin: str | None = None,
        chainType: str | None = None,
        start_time: int | None = None,
        end_time: int | None = None,
        cursor: str | None = None,
        limit: int | None = None,
    ) -> dict:
        """Список депозитов.

        https://bybit-exchange.github.io/docs/v5/asset/deposit
        """
        url = self._BASE_URL + "/v5/asset/deposit/query-record"
        params = {
            "coin": coin,
            "chainType": chainType,
            "startTime": start_time,
            "endTime": end_time,
            "cursor": cursor,
            "limit": limit,
        }

        return self._make_request("GET", url, True, params=params)

    def withdraw_list(
        self,
        coin: str | None = None,
        withdrawType: int | None = None,
        start_time: int | None = None,
        end_time: int | None = None,
        cursor: str | None = None,
        limit: int | None = None,
    ) -> dict:
        """Список выводов.

        https://bybit-exchange.github.io/docs/v5/asset/withdraw
        """
        url = self._BASE_URL + "/v5/asset/withdraw/query-record"
        params = {
            "coin": coin,
            "withdrawType": withdrawType,
            "startTime": start_time,
            "endTime": end_time,
            "cursor": cursor,
            "limit": limit,
        }

        return self._make_request("GET", url, True, params=params)

    def internal_transfer(
        self,
        transferId: str,
        coin: str,
        amount: str,
        fromAccountType: str,
        toAccountType: str,
    ) -> dict:
        """Внутренний перевод между счетами.

        https://bybit-exchange.github.io/docs/v5/asset/transfer
        """
        url = self._BASE_URL + "/v5/asset/transfer/inter-transfer"
        params = {
            "transferId": transferId,
            "coin": coin,
            "amount": amount,
            "fromAccountType": fromAccountType,
            "toAccountType": toAccountType,
        }

        return self._make_request("POST", url, True, params=params)

    def query_internal_transfer(
        self,
        transferId: str | None = None,
        cursor: str | None = None,
        limit: int | None = None,
    ) -> dict:
        """Запрос статуса внутренних переводов.

        https://bybit-exchange.github.io/docs/v5/asset/transfer
        """
        url = self._BASE_URL + "/v5/asset/transfer/query-inter-transfer-list"
        params = {"transferId": transferId, "cursor": cursor, "limit": limit}

        return self._make_request("GET", url, True, params=params)

    def withdraw(
        self,
        coin: str,
        chain: str,
        address: str,
        amount: str,
        tag: str | None = None,
        requestId: str | None = None,
        timestamp: int | None = None,
        forceChain: int | None = None,
        accountType: str | None = None,
    ) -> dict:
        """Создание заявки на вывод.

        https://bybit-exchange.github.io/docs/v5/asset/withdraw
        """
        url = self._BASE_URL + "/v5/asset/withdraw/create"
        params = {
            "coin": coin,
            "chain": chain,
            "address": address,
            "amount": amount,
            "tag": tag,
            "requestId": requestId,
            "timestamp": timestamp,
            "forceChain": forceChain,
            "accountType": accountType,
        }

        return self._make_request("POST", url, True, params=params)

    def cancel_withdraw(self, withdrawId: str) -> dict:
        """Отмена вывода.

        https://bybit-exchange.github.io/docs/v5/asset/withdraw
        """
        url = self._BASE_URL + "/v5/asset/withdraw/cancel"
        params = {"withdrawId": withdrawId}

        return self._make_request("POST", url, True, params=params)

    def deposit_address(
        self, coin: str, chainType: str | None = None, accountType: str | None = None
    ) -> dict:
        """Адрес депозита.

        https://bybit-exchange.github.io/docs/v5/asset/deposit-address
        """
        url = self._BASE_URL + "/v5/asset/deposit/query-address"
        params = {"coin": coin, "chainType": chainType, "accountType": accountType}

        return self._make_request("GET", url, True, params=params)

    # ==========================================================================
    # SPOT MARGIN (Private)
    # ==========================================================================

    def spot_margin_switch(self, spotMarginMode: int) -> dict:
        """Вкл/выкл Spot Margin Trading.

        https://bybit-exchange.github.io/docs/v5/spot-margin/switch
        """
        url = self._BASE_URL + "/v5/spot-margin/switch"
        params = {"spotMarginMode": spotMarginMode}

        return self._make_request("POST", url, True, params=params)

    def spot_margin_borrow_history(
        self,
        coin: str | None = None,
        start_time: int | None = None,
        end_time: int | None = None,
        limit: int | None = None,
        cursor: str | None = None,
    ) -> dict:
        """История заимствований (Spot Margin).

        https://bybit-exchange.github.io/docs/v5/spot-margin/borrow-history
        """
        url = self._BASE_URL + "/v5/spot-margin/borrow-history"
        params = {
            "coin": coin,
            "startTime": start_time,
            "endTime": end_time,
            "limit": limit,
            "cursor": cursor,
        }

        return self._make_request("GET", url, True, params=params)

    # ==========================================================================
    # OPTIONS (Public/Private) — базовые рыночные данные
    # ==========================================================================

    def options_delivery_price(self, symbol: str) -> dict:
        """Delivery price для опциона.

        https://bybit-exchange.github.io/docs/v5/market/delivery-price
        """
        url = self._BASE_URL + "/v5/market/delivery-price"
        params = {"category": "option", "symbol": symbol}

        return self._make_request("GET", url, params=params)

    # ==========================================================================
    # COPY TRADING / BROKER / LOAN
    # ==========================================================================

    def broker_earnings(
        self,
        start_time: int | None = None,
        end_time: int | None = None,
        cursor: str | None = None,
        limit: int | None = None,
    ) -> dict:
        """Доходы брокера (если аккаунт поддерживает).

        https://bybit-exchange.github.io/docs/v5/broker/...
        """
        url = self._BASE_URL + "/v5/broker/earnings"
        params = {"startTime": start_time, "endTime": end_time, "cursor": cursor, "limit": limit}

        return self._make_request("GET", url, True, params=params)

    def institutional_loan_list(
        self, status: str | None = None, cursor: str | None = None, limit: int | None = None
    ) -> dict:
        """Список займов (Institutional Loan).

        https://bybit-exchange.github.io/docs/v5/loan/...
        """
        url = self._BASE_URL + "/v5/loan/query-list"
        params = {"status": status, "cursor": cursor, "limit": limit}

        return self._make_request("GET", url, True, params=params)



"""Модуль, который предоставляет дополнительные функции, которые нужны для внутренного использования в библиотеке."""

__all__ = [
    "dict_to_query_string",
    "generate_hmac_sha256_signature",
    "sort_params_by_alphabetical_order",
    "filter_params",
    "batched_list",
]

import base64
import hashlib
import hmac
import json
from collections.abc import Iterable
from typing import Literal
from urllib.parse import urlencode


def filter_params(params: dict) -> dict:
    """Фильтрует параметры запроса, удаляя None-значения.

    Параметры:
        params (`dict`): Словарь параметров запроса.

    Возвращает:
        `dict`: Отфильтрованный словарь параметров запроса.
    """
    return {k: v for k, v in params.items() if v is not None}


def sort_params_by_alphabetical_order(params: dict) -> dict:
    """Сортирует параметры запроса по алфавиту.

    Параметры:
        params (`dict`): Словарь параметров запроса.

    Возвращает:
        `dict`: Отсортированный словарь параметров запроса.
    """
    return dict(sorted(params.items()))


def dict_to_query_string(params: dict) -> str:
    """Преобразует словарь параметров в query string для URL.

    - Списки и словари автоматически сериализуются в JSON.
    - Используется стандартная urlencode кодировка.

    Параметры:
        params (`dict`): Словарь параметров запроса.

    Возвращает:
        `str`: Строка параметров, готовая для использования в URL.
    """
    processed = {
        k: json.dumps(v, separators=(",", ":")) if isinstance(v, list | dict) else v
        for k, v in params.items()
    }
    return urlencode(processed, doseq=True)


def generate_hmac_sha256_signature(
    secret_key: str,
    payload: str,
    encoding: Literal["hex", "base64"] = "hex",
) -> str:
    """Генерирует HMAC-SHA256 подпись.

    encoding:
        - "hex" → шестнадцатеричная строка
        - "base64" → base64-строка
    """
    digest = hmac.new(secret_key.encode("utf-8"), payload.encode("utf-8"), hashlib.sha256).digest()
    if encoding == "hex":
        return digest.hex()
    elif encoding == "base64":
        return base64.b64encode(digest).decode()
    else:
        raise ValueError("encoding must be 'hex' or 'base64'")


def batched_list[T](iterable: Iterable[T], n: int) -> list[list[T]]:
    """Разбивает последовательность на чанки фиксированного размера.

    Всегда возвращает список списков (list[list[T]]).
    """
    if n <= 0:
        raise ValueError("n must be greater than 0")

    result: list[list[T]] = []
    batch: list[T] = []
    for item in iterable:
        batch.append(item)
        if len(batch) == n:
            result.append(batch)
            batch = []
    if batch:
        result.append(batch)
    return result



"""Пакет, содержащий реализации синхронных и асинхронных клиентов и менеджеров для работы с биржей Binance."""

__all__ = [
    "Adapter",
    "Client",
    "UniClient",
    "UniWebsocketManager",
    "UserWebsocket",
    "WebsocketManager",
]

from .adapter import Adapter
from .sync import Client, UniClient, UniWebsocketManager, UserWebsocket, WebsocketManager



"""Модуль, который предоставляет мапперы для унифицированных клиентов и вебсокет-менеджеров."""

__all__ = [
    "get_uni_client",
    "get_uni_websocket_manager",
]

from typing import Literal, overload

from ._abc import IUniClient, IUniWebsocketManager
from ._abc.asyncio import IUniClient as IUniAioClient
from ._abc.asyncio import IUniWebsocketManager as IUniAioWebsocketManager
from .binance import UniClient as BinanceUniClient
from .binance import UniWebsocketManager as BinanceUniWebsocketManager
from .binance.asyncio import UniClient as BinanceUniAioClient
from .binance.asyncio import UniWebsocketManager as BinanceUniAioWebsocketManager
from .bitget.asyncio import UniClient as BitgetUniAioClient
from .bitget.asyncio import UniWebsocketManager as BitgetUniAioWebsocketManager
from .enums import Exchange
from .exceptions import NotSupported

_UNI_CLIENT_MAPPER: dict[Exchange, type[IUniClient]] = {Exchange.BINANCE: BinanceUniClient}
"""Маппер, который связывает биржу и синхронную реализацию унифицированного клиента."""

_UNI_AIO_CLIENT_MAPPER: dict[Exchange, type[IUniAioClient]] = {
    Exchange.BINANCE: BinanceUniAioClient,
    Exchange.BITGET: BitgetUniAioClient,
}
"""Маппер, который связывает биржу и асинхронную реализацию унифицированного клиента."""

_UNI_WS_MANAGER_MAPPER: dict[Exchange, type[IUniWebsocketManager]] = {
    Exchange.BINANCE: BinanceUniWebsocketManager
}
"""Маппер, который связывает биржу и синхронную реализацию унифицированного вебсокет-менеджера."""

_UNI_AIO_WS_MANAGER_MAPPER: dict[Exchange, type[IUniAioWebsocketManager]] = {
    Exchange.BINANCE: BinanceUniAioWebsocketManager,
    Exchange.BITGET: BitgetUniAioWebsocketManager,
}
"""Маппер, который связывает биржу и асинхронную реализацию унифицированного вебсокет-менеджера."""


@overload
def get_uni_client(exchange: Exchange) -> type[IUniClient]: ...


@overload
def get_uni_client(exchange: Exchange, is_async: Literal[True]) -> type[IUniAioClient]: ...


@overload
def get_uni_client(exchange: Exchange, is_async: Literal[False]) -> type[IUniClient]: ...


def get_uni_client(
    exchange: Exchange, is_async: bool = False
) -> type[IUniClient] | type[IUniAioClient]:
    """Возвращает унифицированный клиент для указанной биржи.

    Параметры:
        exchange (`Exchange`): Биржа.
        is_async (`bool`): Если True, возвращает асинхронный клиент. Иначе - синхронный. По умолчанию False.

    Возвращает:
        `type[IUniClient] | type[IUniAioClient]`: Унифицированный клиент для указанной биржи.
    """
    try:
        if is_async:
            return _UNI_AIO_CLIENT_MAPPER[exchange]
        else:
            return _UNI_CLIENT_MAPPER[exchange]
    except KeyError as e:
        raise NotSupported(f"Unsupported exchange: {exchange}") from e


@overload
def get_uni_websocket_manager(exchange: Exchange) -> type[IUniWebsocketManager]: ...


@overload
def get_uni_websocket_manager(
    exchange: Exchange, is_async: Literal[True]
) -> type[IUniAioWebsocketManager]: ...


@overload
def get_uni_websocket_manager(
    exchange: Exchange, is_async: Literal[False]
) -> type[IUniWebsocketManager]: ...


def get_uni_websocket_manager(
    exchange: Exchange, is_async: bool = True
) -> type[IUniWebsocketManager] | type[IUniAioWebsocketManager]:
    """Возвращает унифицированный вебсокет-менеджер для указанной биржи.

    Параметры:
        exchange (`Exchange`): Биржа.
        is_async (`bool`): Если True, возвращает асинхронный вебсокет-менеджер. Иначе - синхронный. По умолчанию True.

    Возвращает:
        `type[IUniWebsocketManager] | type[IUniAioWebsocketManager]`: Унифицированный вебсокет-менеджер для указанной биржи.
    """
    try:
        if is_async:
            return _UNI_AIO_WS_MANAGER_MAPPER[exchange]
        else:
            return _UNI_WS_MANAGER_MAPPER[exchange]
    except KeyError as e:
        raise NotSupported(f"Unsupported exchange: {exchange}") from e



__all__ = ["UniClient"]

from functools import cached_property

from unicex._abc import IUniClient
from unicex.enums import Exchange, Timeframe
from unicex.exceptions import NotSupported
from unicex.types import KlineDict, TickerDailyDict

from ..adapter import Adapter
from .client import Client


class UniClient(IUniClient[Client]):
    """Унифицированный клиент для работы с Binance API."""

    @cached_property
    def adapter(self) -> Adapter:
        """Возвращает реализацию адаптера для Binance.

        Возвращает:
            BinanceAdapter: Реализация адаптера для Binance.
        """
        return Adapter()

    @property
    def _client_cls(self) -> type[Client]:
        """Возвращает класс клиента для Binance.

        Возвращает:
            type[BinanceClient]: Класс клиента для Binance.
        """
        return Client

    def tickers(self, only_usdt: bool = True) -> list[str]:
        """Возвращает список тикеров.

        Параметры:
            only_usdt (bool): Если True, возвращает только тикеры в паре к USDT.

        Возвращает:
            list[str]: Список тикеров.
        """
        raw_data = self._client.ticker_price()
        return self.adapter.tickers(raw_data=raw_data, only_usdt=only_usdt)  # type: ignore | raw_data is list[dict] if symbol param is not ommited

    def futures_tickers(self, only_usdt: bool = True) -> list[str]:
        """Возвращает список тикеров.

        Параметры:
            only_usdt (bool): Если True, возвращает только тикеры в паре к USDT.

        Возвращает:
            list[str]: Список тикеров.
        """
        raw_data = self._client.futures_ticker_price()
        return self.adapter.tickers(raw_data=raw_data, only_usdt=only_usdt)  # type: ignore | raw_data is list[dict] if symbol param is not ommited

    def ticker_24h(self) -> dict[str, TickerDailyDict]:
        """Возвращает статистику за последние 24 часа для каждого тикера.

        Возвращает:
            dict[str, TickerDailyDict]: Словарь с статистикой за последние 24 часа для каждого тикера.
        """
        raw_data = self._client.ticker_24h()
        return self.adapter.ticker_24h(raw_data=raw_data)  # type: ignore | raw_data is list[dict] if symbol param is not ommited

    def futures_ticker_24h(self) -> dict[str, TickerDailyDict]:
        """Возвращает статистику за последние 24 часа для каждого тикера.

        Возвращает:
            dict[str, TickerDailyDict]: Словарь с статистикой за последние 24 часа для каждого тикера.
        """
        raw_data = self._client.futures_ticker_24h()
        return self.adapter.futures_ticker_24h(raw_data=raw_data)  # type: ignore | raw_data is list[dict] if symbol param is not ommited

    def last_price(self) -> dict[str, float]:
        """Возвращает последнюю цену для каждого тикера.

        Возвращает:
            dict[str, float]: Словарь с последними ценами для каждого тикера.
        """
        raw_data = self._client.ticker_price()
        return self.adapter.last_price(raw_data)  # type: ignore | raw_data is list[dict] if symbol param is not ommited

    def futures_last_price(self) -> dict[str, float]:
        """Возвращает последнюю цену для каждого тикера.

        Возвращает:
            dict[str, float]: Словарь с последними ценами для каждого тикера.
        """
        raw_data = self._client.futures_ticker_price()
        return self.adapter.futures_last_price(raw_data)  # type: ignore | raw_data is list[dict] if symbol param is not ommited

    def klines(
        self,
        symbol: str,
        interval: Timeframe,
        limit: int | None = None,
        start_time: int | None = None,
        end_time: int | None = None,
    ) -> list[KlineDict]:
        """Возвращает список свечей для тикера.

        Параметры:
            symbol (str): Название тикера.
            limit (int | None): Количество свечей.
            interval (Timeframe): Таймфрейм свечей.
            start_time (int | None): Время начала периода в миллисекундах.
            end_time (int | None): Время окончания периода в миллисекундах.

        Возвращает:
            list[KlineDict]: Список свечей для тикера.
        """
        raw_data = self._client.klines(
            symbol=symbol,
            interval=interval.to_exchange_format(Exchange.BINANCE),  # type: ignore
            limit=limit,
            start_time=start_time,
            end_time=end_time,
        )
        return self.adapter.klines(raw_data)

    def futures_klines(
        self,
        symbol: str,
        interval: Timeframe | str,
        limit: int | None = None,
        start_time: int | None = None,
        end_time: int | None = None,
    ) -> list[KlineDict]:
        """Возвращает список свечей для тикера.

        Параметры:
            symbol (str): Название тикера.
            limit (int | None): Количество свечей.
            interval (Timeframe): Таймфрейм свечей.
            start_time (int | None): Время начала периода в миллисекундах.
            end_time (int | None): Время окончания периода в миллисекундах.

        Возвращает:
            list[KlineDict]: Список свечей для тикера.
        """
        raw_data = self._client.futures_klines(
            symbol=symbol,
            interval=interval.to_exchange_format(Exchange.BINANCE),  # type: ignore
            limit=limit,
            start_time=start_time,
            end_time=end_time,
        )
        return self.adapter.futures_klines(raw_data)

    def funding_rate(self) -> dict[str, float]:
        """Возвращает ставку финансирования для всех тикеров.

        Возвращает:
            dict[str, float]: Ставка финансирования для каждого тикера.
        """
        raw_data = self._client.futures_mark_price()
        return self.adapter.funding_rate(raw_data)  # type: ignore | raw_data is list[dict] if symbol param is not ommited

    def open_interest(self, symbol: str) -> float:
        """Возвращает объем открытого интереса для тикера или всех тикеров,
        если тикер не указан.

        Параметры:
            symbol (str | None): Название тикера (Опционально).

        Возвращает:
            float: Объем открытых позиций в монетах.
        """
        if not symbol:
            raise NotSupported("Open interest for all symbols is not supported for binance")

        raw_data = self._client.open_interest(symbol)
        return self.adapter.open_interest(raw_data)  # type: ignore | binance supports only single ticker open interest



__all__ = ["Websocket"]

import threading
import time
from collections import deque
from collections.abc import Callable
from typing import Any

import orjson
from loguru import logger as _logger
from websocket import WebSocket, WebSocketApp

from unicex.exceptions import QueueOverflowError
from unicex.types import LoggerLike


class Websocket:
    """Базовый класс синхронного вебсокета."""

    MAX_QUEUE_SIZE: int = 100
    """Максимальная длина очереди."""

    def __init__(
        self,
        callback: Callable[[Any], None],
        url: str,
        subscription_messages: list[dict] | list[str] | None = None,
        ping_interval: int | float | None = 10,
        ping_message: str | None = None,
        pong_message: str | None = None,
        no_message_reconnect_timeout: int | float | None = 60,
        reconnect_timeout: int | float | None = 5,
        worker_count: int = 2,
        logger: LoggerLike | None = None,
        **kwargs: Any,  # Не дадим сломаться, если юзер передал ненужные аргументы
    ) -> None:
        """Инициализация вебсокета.

        Параметры:
            callback (`Callable[[Any], None]`): Обработчик входящих сообщений.
            url (`str`): URL вебсокета.
            subscription_messages (`list[dict] | list[str] | None`): Сообщения для подписки после подключения.
            ping_interval (`int | float | None`): Интервал отправки ping, сек.
            ping_message (`str | None`): Сообщение для ping (если не указано — используется ping‑frame).
            pong_message (`str | None`): Сообщение для pong (если не указано — используется pong‑frame).
            no_message_reconnect_timeout (`int | float | None`): Таймаут без сообщений до рестарта, сек.
            reconnect_timeout (`int | float | None`): Пауза перед переподключением, сек.
            worker_count (`int`): Количество рабочих потоков.
            logger (`LoggerLike | None`): Логгер для записи логов.
        """
        self._callback = callback
        self._subscription_messages = subscription_messages or []
        self._ping_interval = ping_interval
        self._ping_message = ping_message
        self._pong_message = pong_message
        self._no_message_reconnect_timeout = no_message_reconnect_timeout
        self._logger = logger or _logger
        self._reconnect_timeout = reconnect_timeout or 0
        self._last_message_time: float | None = None
        self._healthcheck_thread: threading.Thread | None = None
        self._ws_thread: threading.Thread | None = None
        self._ws = WebSocketApp(
            url=url,
            on_open=self._on_open,
            on_message=self._on_message,
            on_error=self._on_error,
            on_close=self._on_close,
        )

        # Очередь сообщений
        self._queue: deque = deque()
        self._queue_lock = threading.Lock()

        # Воркеры
        self._worker_count = worker_count
        self._workers: list[threading.Thread] = []

        self._running = False

    def start(self) -> None:
        """Запускает вебсокет в отдельном потоке."""
        # Проверяем что вебсокет еще не запущен
        if self._running:
            raise RuntimeError("Websocket is already running")
        self._running = True

        self._logger.info("Starting websocket")

        # Запускаем вебсокет
        self._ws_thread = threading.Thread(
            target=self._ws.run_forever,
            name="WebsocketThread",
            kwargs=self._generate_ws_kwargs(),
            daemon=True,
        )
        self._ws_thread.start()

        # Инициализируем время последнего сообщения
        self._last_message_time = time.monotonic()

        # Запускаем поток для проверки времени последнего сообщения
        self._healthcheck_thread = threading.Thread(
            target=self._healthcheck_task,
            name="WebsocketHealthcheckThread",
            daemon=True,
        )
        self._healthcheck_thread.start()

        # Запускаем воркеров
        for i in range(self._worker_count):
            worker = threading.Thread(
                target=self._worker,
                name=f"WebsocketWorkerThread-{i}",
                daemon=True,
            )
            self._workers.append(worker)
            worker.start()

    def stop(self) -> None:
        """Останавливает вебсокет и поток."""
        self._logger.info("Stopping websocket")
        self._running = False

        # Останавливаем поток проверки последнего времени сообщения
        try:
            if isinstance(self._healthcheck_thread, threading.Thread):
                self._healthcheck_thread.join(timeout=1)
        except Exception as e:
            self._logger.error(f"Error stopping healthcheck thread: {e}")

        # Останавилваем вебсокет
        try:
            if self._ws:
                self._ws.close()  # отправляем "close frame"
        except Exception as e:
            self._logger.error(f"Error closing websocket: {e}")

        # Ждем остановки вебсокета
        try:
            if self._ws_thread and self._ws_thread.is_alive():
                self._ws_thread.join(timeout=5)  # ждём завершения потока
                self._ws_thread = None
        except Exception as e:
            self._logger.error(f"Error stopping websocket thread: {e}")

        # Ждем остановки воркеров
        for worker in self._workers:
            try:
                if worker.is_alive():
                    worker.join(timeout=1)
            except Exception as e:
                self._logger.error(f"Error stopping worker: {e}")
        self._workers.clear()

    def restart(self) -> None:
        """Перезапускает вебсокет."""
        self.stop()
        time.sleep(self._reconnect_timeout)
        self.start()

    def _worker(self) -> None:
        """Цикл обработки сообщений из очереди."""
        while self._running:
            try:
                with self._queue_lock:
                    if not self._queue:
                        time.sleep(0.01)
                        continue
                    message = self._queue.popleft()

                self._callback(message)
            except Exception as e:
                self._logger.error(f"Error in worker: {e}")
            time.sleep(0.01)  # чтобы не крутиться на пустой очереди

    def _generate_ws_kwargs(self) -> dict:
        """Генерирует аргументы для запуска вебсокета."""
        ws_kwargs = {}
        if self._ping_interval:
            ws_kwargs["ping_interval"] = self._ping_interval
        if self._ping_message:
            ws_kwargs["ping_payload"] = self._ping_message
        return ws_kwargs

    def _on_open(self, ws: WebSocket) -> None:
        """Обработчик события открытия вебсокета."""
        self._logger.info("Websocket opened")
        for subscription_message in self._subscription_messages:
            if isinstance(subscription_message, dict):
                subscription_message = orjson.dumps(subscription_message)  # noqa: PLW2901
            ws.send(subscription_message)

    def _on_message(self, _: WebSocket, message: str) -> None:
        """Обработчик события получения сообщения."""
        try:
            message = orjson.loads(message)
        except orjson.JSONDecodeError as e:
            if message in ["ping", "pong"]:
                self._logger.debug(f"{self} Received ping message: {message}")
            else:
                self._logger.error(f"Failed to decode JSON message: {message}, error: {e}")

        self._last_message_time = time.monotonic()

        with self._queue_lock:
            if len(self._queue) >= self.MAX_QUEUE_SIZE:
                raise QueueOverflowError("Message queue is overflow")
            self._queue.append(message)

    def _on_error(self, _: WebSocket, error: Exception) -> None:
        """Обработчик события ошибки вебсокета."""
        self._logger.error(f"Websocket error: {error}")
        self.restart()

    def _on_close(self, _: WebSocket, status_code: int, reason: str) -> None:
        """Обработчик события закрытия вебсокета."""
        self._logger.info(f"Websocket closed with status code {status_code} and reason {reason}")

    def _on_ping(self, ws: WebSocket, message: str) -> None:
        """Обработчик события получения ping."""
        self._logger.info(f"Websocket received ping: {message}")
        if self._pong_message:
            ws.pong(self._pong_message)
        else:
            ws.pong()

    def _healthcheck_task(self) -> None:
        """Проверяет работоспособность по времени последнего сообщения."""
        if not self._no_message_reconnect_timeout:
            return

        while self._running:
            try:
                if time.monotonic() - self._last_message_time > self._no_message_reconnect_timeout:  # type: ignore
                    self._logger.error("Websocket no message timeout triggered")
                    self.restart()
            except Exception as e:
                self._logger.error(f"Error checking websocket health: {e}")
            time.sleep(1)



"""Модуль, который предоставляет дополнительные функции, которые могут пригодиться в работе."""

__all__ = [
    "percent_greater",
    "percent_less",
    "TimeoutTracker",
    "generate_ex_link",
    "generate_tv_link",
    "generate_cg_link",
]

import time

from .enums import Exchange, MarketType
from .exceptions import NotSupported


def percent_greater(higher: float, lower: float) -> float:
    """Возвращает на сколько процентов `higher` больше `lower`.

    Можно воспринимать полученное значение как если вести линейку на tradingview.com от меньшего значения к большему.

    Например:
        ```python
        percent_greater(120, 100)
        >> 20.0
        ```

    Возвращает:
        `float`: На сколько процентов `higher` больше `lower`.
    """
    if lower == 0:
        return float("inf")
    return (higher / lower - 1) * 100


def percent_less(higher: float, lower: float) -> float:
    """Возвращает на сколько процентов `lower` меньше `higher`.

    Можно воспринимать полученное значение как если вести линейку на tradingview.com от большего значения к меньшему.

    Например:
        ```python
        percent_less(120, 100)
        >> 16.67777777777777
        ```

    Возвращает:
        `float`: На сколько процентов `lower` меньше `higher`.
    """
    if lower == 0:
        return float("inf")
    return (1 - lower / higher) * 100


class TimeoutTracker[T]:
    """Универсальный менеджер для управления таймаутами любых объектов.
    Позволяет временно блокировать объекты на заданный промежуток времени.
    """

    def __init__(self) -> None:
        """Инициализирует пустой словарь для отслеживания заблокированных объектов."""
        self._blocked_items: dict[T, float] = {}

    def is_blocked(self, item: T) -> bool:
        """Проверяет, находится ли объект в состоянии блокировки.
        Если срок блокировки истёк, удаляет объект из списка.

        Параметры:
            item (`T`): Объект, который нужно проверить.

        Возвращает:
            `bool`: True, если объект заблокирован, иначе False.
        """
        if item in self._blocked_items:
            if time.time() < self._blocked_items[item]:
                return True
            else:
                del self._blocked_items[item]
        return False

    def block(self, item: T, duration: int) -> None:
        """Блокирует объект на указанное количество секунд.

        Параметры:
            item (`T`): Объект, который нужно заблокировать.
            duration (`int`): Длительность блокировки в секундах.
        """
        self._blocked_items[item] = time.time() + duration


def generate_ex_link(exchange: Exchange, market_type: MarketType, symbol: str):
    """Генерирует ссылку на биржу.

    Параметры:
        exchange (`Exchange`): Биржа.
        market_type (`MarketType`): Тип рынка.
        symbol (`str`): Символ.

    Возвращает:
        `str`: Ссылка на биржу.
    """
    ticker = symbol.removesuffix("USDT").removesuffix("_USDT").removesuffix("-USDT")
    if exchange == Exchange.BINANCE:
        if market_type == MarketType.FUTURES:
            return f"https://www.binance.com/en/futures/{symbol}"
        else:
            return f"https://www.binance.com/en/trade/{ticker}_USDT?type=spot"
    elif exchange == Exchange.BYBIT:
        if market_type == MarketType.FUTURES:
            return f"https://www.bybit.com/trade/usdt/{symbol}"
        else:
            return f"https://www.bybit.com/en/trade/spot/{ticker}/USDT"
    elif exchange == Exchange.BITGET:
        if market_type == MarketType.FUTURES:
            return f"https://www.bitget.com/ru/futures/usdt/{symbol}"
        else:
            return f"https://www.bitget.com/ru/spot/{symbol}"
    else:
        raise NotSupported(f"Exchange {exchange} is not supported")


def generate_tv_link(exchange: Exchange, market_type: MarketType, symbol: str) -> str:
    """Генерирует ссылку для TradingView.

    Параметры:
        exchange (`Exchange`): Биржа.
        market_type (`MarketType`): Тип рынка.
        symbol (`str`): Символ.

    Возвращает:
        `str`: Ссылка для TradingView.
    """
    if market_type == MarketType.FUTURES:
        return f"https://www.tradingview.com/chart/?symbol={exchange}:{symbol}.P"
    elif market_type == MarketType.SPOT:
        return f"https://www.tradingview.com/chart/?symbol={exchange}:{symbol}"
    else:
        raise NotSupported(f"Unsupported market type: {market_type}")


def generate_cg_link(exchange: Exchange, market_type: MarketType, symbol: str) -> str:
    """Генерирует ссылку для CoinGlass.

    Параметры:
        exchange (`Exchange`): Биржа.
        market_type (`MarketType`): Тип рынка.
        symbol (`str`): Символ.

    Возвращает:
        `str`: Ссылка для CoinGlass.
    """
    if market_type == MarketType.FUTURES:
        if exchange == Exchange.BITGET:
            # https://www.coinglass.com/tv/ru/Bitget_ETHUSDT_UMCBL
            return f"https://www.coinglass.com/tv/ru/{exchange.capitalize()}_{symbol}_UMCBL"
        else:
            # Стандартный вид ссылки (Подходит для BYBIT и BINANCE)
            return f"https://www.coinglass.com/tv/ru/{exchange.capitalize()}_{symbol}"
    elif market_type == MarketType.SPOT:
        if exchange == Exchange.BITGET:
            # Для спота нет ссылки на койнгласс
            return f"https://www.coinglass.com/tv/ru/{exchange.capitalize()}_{symbol}_UMCBL"
        else:
            # Стандартный вид ссылки (Подходит для BYBIT и BINANCE)
            return f"https://www.coinglass.com/tv/ru/SPOT_{exchange.capitalize()}_{symbol}"
    else:
        raise NotSupported(f"Market type {market_type} is not supported")



__all__ = ["UserWebsocket"]

import threading
import time
from collections.abc import Callable
from typing import Any

from loguru import logger as _logger

from unicex._base import Websocket
from unicex.exceptions import NotSupported
from unicex.types import LoggerLike

from .._mixins import UserWebsocketMixin
from ..types import AccountType
from .client import Client


class UserWebsocket(UserWebsocketMixin):
    """Пользовательский вебсокет Binance с авто‑продлением listenKey.

    Поддержка типов аккаунта: "SPOT" и "FUTURES" (USDT‑M фьючерсы).
    """

    def __init__(
        self,
        callback: Callable[[Any], None],
        client: Client,
        type: AccountType,
        logger: LoggerLike | None = None,
        **kwargs: Any,  # Не дадим сломаться, если юзер передал ненужные аргументы
    ) -> None:
        """Инициализирует пользовательский вебсокет.

        Параметры:
            callback (`Callable`): Функция обратного вызова, которая принимает сообщение с вебсокета.
            client (`Client`): Авторизованный клиент Binance.
            type (`AccountType`): Тип аккаунта ("SPOT" | "FUTURES").
            logger (`LoggerLike | None`): Логгер для записи логов.
        """
        self._callback = callback
        self._client = client
        self._type = type

        self._listen_key: str | None = None
        self._ws: Websocket | None = None
        self._keepalive_thread: threading.Thread | None = None

        self._logger = logger or _logger

        self._running = False

    def start(self) -> None:
        """Запускает пользовательский стрим с автопродлением listenKey."""
        self._running = True
        self._listen_key = self._create_listen_key()
        self._start_ws(self._create_ws_url(self._type, self._listen_key))  # type: ignore

        # Фоновое продление ключа прослушивания
        self._keepalive_thread = threading.Thread(
            target=self._keepalive_loop, name="WebsocketKeepaliveThread", daemon=True
        )
        self._keepalive_thread.start()

    def stop(self) -> None:
        """Останавливает стрим и закрывает listenKey."""
        self._running = False

        # Останавливаем вебсокет
        try:
            if isinstance(self._ws, Websocket):
                self._ws.stop()
        except Exception as e:
            self._logger.error(f"Error stopping WebSocket: {e}")

        # Ожидаем завершения фонового продления ключа прослушивания
        if isinstance(self._keepalive_thread, threading.Thread):
            self._keepalive_thread.join()

        # Закрываем ключ прослушивания
        try:
            if self._listen_key:
                self._close_listen_key(self._listen_key)
        except Exception as e:
            self._logger.error(f"Error closing listenKey: {e}")
        finally:
            self._listen_key = None

        self._logger.info("User websocket stopped")

    def restart(self) -> None:
        """Перезапускает WebSocket для User Data Stream."""
        self.stop()
        self.start()

    def _start_ws(self, ws_url: str) -> None:
        """Запускает WebSocket для User Data Stream."""
        self._ws = Websocket(callback=self._callback, url=ws_url, no_message_reconnect_timeout=None)
        self._ws.start()
        self._logger.info(f"User websocket started: ...{ws_url[-5:]}")

    def _keepalive_loop(self) -> None:
        """Фоновый цикл продления listenKey и восстановления сессии при необходимости."""
        while self._running:
            try:
                if self._type == "FUTURES":
                    response = self._renew_listen_key()
                    listen_key = response.get("listenKey")
                    if not listen_key:
                        raise RuntimeError(f"Can not renew listenKey: {response}")

                    if listen_key != self._listen_key:
                        self._logger.info(
                            f"Listen key changed: {self._listen_key} -> {listen_key}. Restarting websocket"
                        )
                        return self.restart()

                elif self._type == "SPOT":
                    self._renew_listen_key()

                else:
                    raise NotSupported(f"Account type '{self._type}' not supported")

            except Exception as e:
                self._logger.error(f"Error while keeping alive: {e}")
                return self.restart()

            # Ждём до следующего продления
            for _ in range(self._RENEW_INTERVAL):
                if self._running:
                    return
                time.sleep(1)

    def _create_listen_key(self) -> str:
        """Создает новый listenKey для User Data Stream в зависимости от типа аккаунта."""
        if self._type == "FUTURES":
            resp = self._client.futures_listen_key()
        elif self._type == "SPOT":
            resp = self._client.listen_key()
        else:
            raise NotSupported(f"Account type '{self._type}' not supported")

        key = resp.get("listenKey")
        if not key:
            raise RuntimeError(f"Can not create listenKey: {resp}")
        return key

    def _renew_listen_key(self) -> dict:
        """Продлевает listenKey. Возвращает новый ключ, если сервер его выдал."""
        if not isinstance(self._listen_key, str):
            raise RuntimeError("listenKey is not a string")
        if self._type == "FUTURES":
            return self._client.futures_renew_listen_key()
        elif self._type == "SPOT":
            return self._client.renew_listen_key(self._listen_key)
        else:
            raise NotSupported(f"Account type '{self._type}' not supported")

    def _close_listen_key(self, listen_key: str) -> None:
        """Закрывает listenKey."""
        if self._type == "FUTURES":
            self._client.futures_close_listen_key()
        elif self._type == "SPOT":
            self._client.close_listen_key(listen_key)
        else:
            raise NotSupported(f"Account type '{self._type}' not supported")



__all__ = ["IUniClient"]

from abc import ABC, abstractmethod
from collections.abc import Sequence
from functools import cached_property
from itertools import batched
from typing import Generic, Self, TypeVar, overload

import requests

from unicex._base import BaseClient
from unicex.enums import Timeframe
from unicex.types import KlineDict, LoggerLike, TickerDailyDict

from ..adapter import IAdapter

TClient = TypeVar("TClient", bound="BaseClient")


class IUniClient(ABC, Generic[TClient]):
    """Интерфейс для реализации синхронного унифицированного клиента."""

    def __init__(
        self,
        api_key: str | None = None,
        api_secret: str | None = None,
        api_passphrase: str | None = None,
        session: requests.Session | None = None,
        logger: LoggerLike | None = None,
        max_retries: int = 3,
        retry_delay: int | float = 0.1,
        proxies: list[str] | None = None,
        timeout: int = 10,
    ) -> None:
        """Инициализация клиента.

        Параметры:
            api_key (str | None): Ключ API для аутентификации.
            api_secret (str | None): Секретный ключ API для аутентификации.
            api_passphrase (`str | None`): Пароль API для аутентификации (Bitget).
            session (requests.Session): Сессия для выполнения HTTP-запросов.
            logger (LoggerLike | None): Логгер для вывода информации.
            max_retries (int): Максимальное количество повторных попыток запроса.
            retry_delay (int | float): Задержка между повторными попытками.
            proxies (list[str] | None): Список HTTP(S) прокси для циклического использования.
            timeout (int): Максимальное время ожидания ответа от сервера.
        """
        self._client: TClient = self._client_cls(
            api_key=api_key,
            api_secret=api_secret,
            api_passphrase=api_passphrase,
            session=session,
            logger=logger,
            max_retries=max_retries,
            retry_delay=retry_delay,
            proxies=proxies,
            timeout=timeout,
        )

    @classmethod
    def from_client(cls, client: TClient) -> Self:
        """Создает UniClient из уже существующего Client.

        Принимает:
            client (`TClient`): Экземпляр Client.

        Возвращает:
            `Self`: Созданный экземпляр клиента.
        """
        instance = cls.__new__(cls)  # создаем пустой объект без вызова __init__
        instance._client = client
        return instance

    def close_connection(self) -> None:
        """Закрывает сессию клиента."""
        self._client.close_connection()

    def __enter__(self) -> Self:
        """Вход в контекст."""
        return self

    def __exit__(self, *_) -> None:
        """Выход из контекста."""
        self.close_connection()

    @property
    def client(self) -> TClient:
        """Возвращает клиент биржи.

        Возвращает:
            `TClient`: Клиент биржи.
        """
        return self._client

    @property
    @abstractmethod
    def _client_cls(self) -> type[TClient]:
        """Возвращает класс клиента для конкретной биржи.

        Возвращает:
            `type[TClient]`: Класс клиента.
        """
        pass

    @cached_property
    @abstractmethod
    def adapter(self) -> IAdapter:
        """Возвращает реализацию адаптера под конкретную биржу.

        Возвращает:
            `IAdapter`: Реализация адаптера.
        """
        pass

    @abstractmethod
    def tickers(self, only_usdt: bool = True) -> list[str]:
        """Возвращает список тикеров.

        Параметры:
            only_usdt (`bool`): Если True, возвращает только тикеры в паре к USDT.

        Возвращает:
            `list[str]`: Список тикеров.
        """
        pass

    def tickers_batched(self, only_usdt: bool = True, batch_size: int = 20) -> list[Sequence[str]]:
        """Возвращает список тикеров в чанках.

        Параметры:
            only_usdt (`bool`): Если True, возвращает только тикеры в паре к USDT.
            batch_size (`int`): Размер чанка.

        Возвращает:
            `list[list[str]]`: Список тикеров в чанках.
        """
        tickers = self.tickers(only_usdt=only_usdt)
        return list(batched(tickers, n=batch_size))

    @abstractmethod
    def futures_tickers(self, only_usdt: bool = True) -> list[str]:
        """Возвращает список тикеров.

        Параметры:
            only_usdt (`bool`): Если True, возвращает только тикеры в паре к USDT.

        Возвращает:
            `list[str]`: Список тикеров.
        """
        pass

    def futures_tickers_batched(
        self, only_usdt: bool = True, batch_size: int = 20
    ) -> list[Sequence[str]]:
        """Возвращает список тикеров в чанках.

        Параметры:
            only_usdt (`bool`): Если True, возвращает только тикеры в паре к USDT.
            batch_size (`int`): Размер чанка.

        Возвращает:
            `list[list[str]]`: Список тикеров в чанках.
        """
        tickers = self.futures_tickers(only_usdt=only_usdt)
        return list(batched(tickers, n=batch_size))

    @abstractmethod
    def last_price(self) -> dict[str, float]:
        """Возвращает последнюю цену для каждого тикера.

        Возвращает:
            `dict[str, float]`: Словарь с последними ценами для каждого тикера.
        """
        pass

    @abstractmethod
    def futures_last_price(self) -> dict[str, float]:
        """Возвращает последнюю цену для каждого тикера.

        Возвращает:
            `dict[str, float]`: Словарь с последними ценами для каждого тикера.
        """
        pass

    @abstractmethod
    def ticker_24h(self) -> dict[str, TickerDailyDict]:
        """Возвращает статистику за последние 24 часа для каждого тикера.

        Возвращает:
            `dict[str, TickerDailyDict]`: Словарь с статистикой за последние 24 часа для каждого тикера.
        """
        pass

    @abstractmethod
    def futures_ticker_24h(self) -> dict[str, TickerDailyDict]:
        """Возвращает статистику за последние 24 часа для каждого тикера.

        Возвращает:
            `dict[str, TickerDailyDict]`: Словарь с статистикой за последние 24 часа для каждого тикера.
        """
        pass

    @abstractmethod
    def klines(
        self,
        symbol: str,
        interval: Timeframe,
        limit: int | None = None,
        start_time: int | None = None,
        end_time: int | None = None,
    ) -> list[KlineDict]:
        """Возвращает список свечей.

        Параметры:
            symbol (`str`): Название тикера.
            interval (`Timeframe`): Таймфрейм свечей.
            limit (`int`): Количество свечей.
            start_time (`int`): Время начала периода в миллисекундах.
            end_time (`int`): Время окончания периода в миллисекундах.

        Возвращает:
            `list[KlineDict]`: Список свечей.
        """
        pass

    @abstractmethod
    def futures_klines(
        self,
        symbol: str,
        interval: Timeframe,
        limit: int | None = None,
        start_time: int | None = None,
        end_time: int | None = None,
    ) -> list[KlineDict]:
        """Возвращает список свечей.

        Параметры:
            symbol (`str`): Название тикера.
            interval (`Timeframe`): Таймфрейм свечей.
            limit (`int`): Количество свечей.
            start_time (`int`): Время начала периода в миллисекундах.
            end_time (`int`): Время окончания периода в миллисекундах.

        Возвращает:
            `list[KlineDict]`: Список свечей.
        """
        pass

    @abstractmethod
    def funding_rate(self) -> dict[str, float]:
        """Возвращает ставку финансирования для всех тикеров.

        Возвращает:
            `dict[str, float]`: Ставка финансирования для каждого тикера.
        """
        pass

    @overload
    def open_interest(self, symbol: str) -> float: ...

    @overload
    def open_interest(self, symbol: None) -> dict[str, float]: ...

    @abstractmethod
    def open_interest(self, symbol: str | None) -> float | dict[str, float]:
        """Возвращает объем открытого интереса для тикера или всех тикеров,
        если тикер не указан.

        Параметры:
            symbol (`str | None`): Название тикера (Опционально).

        Возвращает:
            `float`: Объем открытых позиций в монетах.
        """
        pass



__all__ = ["WebsocketManager"]


from collections.abc import Awaitable, Callable, Sequence
from typing import Any

from unicex._base.asyncio import Websocket
from unicex.exceptions import NotAuthorized

from .._mixins import WebsocketManagerMixin
from ..types import (
    BookDepthLevels,
    ContinuousContractType,
    FuturesTimeframe,
    MarkPriceUpdateSpeed,
    RollingWindowSize,
    SpotTimeframe,
)
from .client import Client
from .user_websocket import UserWebsocket

type CallbackType = Callable[[Any], Awaitable[None]]


class WebsocketManager(WebsocketManagerMixin):
    """Менеджер асинхронных вебсокетов для Binance."""

    def __init__(self, client: Client | None = None, **ws_kwargs: Any) -> None:
        """Инициализирует менеджер вебсокетов для Binance.

        Параметры:
            client (`Client | None`): Клиент для выполнения запросов. Нужен, чтобы открыть приватные вебсокеты.
            ws_kwargs (`dict[str, Any]`): Дополнительные аргументы, котоыре прокидываются в `Websocket`.
        """
        self.client = client
        self._ws_kwargs = ws_kwargs

    def trade(
        self,
        callback: CallbackType,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        """Создает вебсокет для получения сделок.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            symbol (`str | None`): Один символ для подписки.
            symbols (`Sequence[str] | None`): Список символов для мультиплекс‑подключения.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        url = self._generate_stream_url(
            type="trade",
            url=self._BASE_SPOT_URL,
            symbol=symbol,
            symbols=symbols,
            require_symbol=True,
        )
        return Websocket(callback=callback, url=url, **self._ws_kwargs)

    def agg_trade(
        self,
        callback: CallbackType,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        """Создает вебсокет для получения агрегированных сделок.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            symbol (`str | None`): Один символ для подписки.
            symbols (`Sequence[str] | None`): Список символов для мультиплекс‑подключения.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        url = self._generate_stream_url(
            type="aggTrade",
            url=self._BASE_SPOT_URL,
            symbol=symbol,
            symbols=symbols,
            require_symbol=True,
        )
        return Websocket(callback=callback, url=url, **self._ws_kwargs)

    def klines(
        self,
        callback: CallbackType,
        interval: SpotTimeframe,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        """Создает вебсокет для получения свечей.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            symbol (`str | None`): Один символ для подписки.
            interval (`SpotTimeframe`): Временной интервал свечей.
            symbols (`Sequence[str] | None`): Список символов для мультиплекс‑подключения.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        url = self._generate_stream_url(
            type=f"kline_{interval}",
            url=self._BASE_SPOT_URL,
            symbol=symbol,
            symbols=symbols,
            require_symbol=True,
        )
        return Websocket(callback=callback, url=url, **self._ws_kwargs)

    def depth_stream(
        self,
        callback: CallbackType,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        """Создает вебсокет для получения событий изменения стакана (без лимита глубины).

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            symbol (`str | None`): Один символ для подписки.
            symbols (`Sequence[str] | None`): Список символов для мультиплекс‑подключения.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        url = self._generate_stream_url(
            type="depth",
            url=self._BASE_SPOT_URL,
            symbol=symbol,
            symbols=symbols,
            require_symbol=True,
        )
        return Websocket(callback=callback, url=url, **self._ws_kwargs)

    def symbol_mini_ticker(
        self,
        callback: CallbackType,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        """Создает вебсокет для мини‑статистики тикера за последние 24 часа.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            symbol (`str | None`): Один символ для подписки.
            symbols (`Sequence[str] | None`): Список символов для мультиплекс‑подключения.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        url = self._generate_stream_url(
            type="miniTicker",
            url=self._BASE_SPOT_URL,
            symbol=symbol,
            symbols=symbols,
            require_symbol=True,
        )
        return Websocket(callback=callback, url=url, **self._ws_kwargs)

    def mini_ticker(self, callback: CallbackType) -> Websocket:
        """Создает вебсокет для получения мини-статистики всех тикеров за последние 24 ч."""
        url = self._generate_stream_url(type="!miniTicker@arr", url=self._BASE_SPOT_URL)
        return Websocket(callback=callback, url=url, **self._ws_kwargs)

    def symbol_ticker(
        self,
        callback: CallbackType,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        """Создает вебсокет для расширенной статистики тикера за последние 24 часа.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            symbol (`str | None`): Один символ для подписки.
            symbols (`Sequence[str] | None`): Список символов для мультиплекс‑подключения.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        url = self._generate_stream_url(
            type="ticker",
            url=self._BASE_SPOT_URL,
            symbol=symbol,
            symbols=symbols,
            require_symbol=True,
        )
        return Websocket(callback=callback, url=url, **self._ws_kwargs)

    def ticker(self, callback: CallbackType) -> Websocket:
        """Создает вебсокет для получения расширенной статистики всех тикеров за последние 24 ч.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        url = self._generate_stream_url(type="!ticker@arr", url=self._BASE_SPOT_URL)
        return Websocket(callback=callback, url=url, **self._ws_kwargs)

    def symbol_rolling_window_ticker(
        self,
        callback: CallbackType,
        window: RollingWindowSize,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        """Создает вебсокет для получения статистики тикера за указанное окно времени.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            window (`RollingWindowSize`): Размер окна статистики.
            symbol (`str | None`): Один символ для подписки.
            symbols (`Sequence[str] | None`): Список символов для мультиплекс‑подключения.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        url = self._generate_stream_url(
            type=f"ticker_{window}",
            url=self._BASE_SPOT_URL,
            symbol=symbol,
            symbols=symbols,
            require_symbol=True,
        )
        return Websocket(callback=callback, url=url, **self._ws_kwargs)

    def rolling_window_ticker(self, callback: CallbackType, window: RollingWindowSize) -> Websocket:
        """Создает вебсокет для получения статистики всех тикеров за указанное окно времени."""
        url = self._generate_stream_url(type=f"!ticker_{window}@arr", url=self._BASE_SPOT_URL)
        return Websocket(callback=callback, url=url, **self._ws_kwargs)

    def avg_price(
        self,
        callback: CallbackType,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        """Создает вебсокет для получения среднего прайса (Average Price).

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            symbol (`str | None`): Один символ для подписки.
            symbols (`Sequence[str] | None`): Список символов для мультиплекс‑подключения.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        url = self._generate_stream_url(
            type="avgPrice",
            url=self._BASE_SPOT_URL,
            symbol=symbol,
            symbols=symbols,
            require_symbol=True,
        )
        return Websocket(callback=callback, url=url, **self._ws_kwargs)

    def book_ticker(
        self,
        callback: CallbackType,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        """Создает вебсокет для получения лучших бид/аск по символам.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            symbol (`str | None`): Один символ для подписки.
            symbols (`Sequence[str] | None`): Список символов для мультиплекс‑подключения.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        url = self._generate_stream_url(
            type="bookTicker",
            url=self._BASE_SPOT_URL,
            symbol=symbol,
            symbols=symbols,
            require_symbol=True,
        )
        return Websocket(callback=callback, url=url, **self._ws_kwargs)

    def book_depth(
        self,
        callback: CallbackType,
        levels: BookDepthLevels,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        """Создает вебсокет для получения стакана глубиной N уровней.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            levels (`BookDepthLevels`): Глубина стакана (уровни).
            symbol (`str | None`): Один символ для подписки.
            symbols (`Sequence[str] | None`): Список символов для мультиплекс‑подключения.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        url = self._generate_stream_url(
            type=f"depth{levels}",
            url=self._BASE_SPOT_URL,
            symbol=symbol,
            symbols=symbols,
            require_symbol=True,
        )
        return Websocket(callback=callback, url=url, **self._ws_kwargs)

    def user_data_stream(self, callback: CallbackType) -> UserWebsocket:
        """Создает вебсокет для получения информации о пользовательских данных."""
        if not self.client or not self.client.is_authorized():
            raise NotAuthorized("You must provide authorized client.")
        return UserWebsocket(callback=callback, client=self.client, type="SPOT", **self._ws_kwargs)

    def multiplex_socket(self, callback: CallbackType, streams: str) -> Websocket:
        """Создает вебсокет для мультиплексирования нескольких стримов в один.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            streams (`str`): Строка с перечислением стримов.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        return Websocket(
            callback=callback, url=self._BASE_SPOT_URL + "?" + streams, **self._ws_kwargs
        )

    def futures_trade(
        self,
        callback: CallbackType,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        """Создает вебсокет для получения сделок.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            symbol (`str | None`): Один символ для подписки.
            symbols (`Sequence[str] | None`): Список символов для мультиплекс‑подключения.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        url = self._generate_stream_url(
            type="trade",
            url=self._BASE_FUTURES_URL,
            symbol=symbol,
            symbols=symbols,
            require_symbol=True,
        )
        return Websocket(callback=callback, url=url, **self._ws_kwargs)

    def futures_agg_trade(
        self,
        callback: CallbackType,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        """Создает вебсокет для получения агрегированных сделок.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            symbol (`str | None`): Один символ для подписки.
            symbols (`Sequence[str] | None`): Список символов для мультиплекс‑подключения.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        url = self._generate_stream_url(
            type="aggTrade",
            url=self._BASE_FUTURES_URL,
            symbol=symbol,
            symbols=symbols,
            require_symbol=True,
        )
        return Websocket(callback=callback, url=url, **self._ws_kwargs)

    def futures_klines(
        self,
        callback: CallbackType,
        interval: FuturesTimeframe,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        """Создает вебсокет для получения свечей.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            interval (`FuturesTimeframe`): Временной интервал свечей.
            symbol (`str | None`): Один символ для подписки.
            symbols (`Sequence[str] | None`): Список символов для мультиплекс‑подключения.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        url = self._generate_stream_url(
            type=f"kline_{interval}",
            url=self._BASE_FUTURES_URL,
            symbol=symbol,
            symbols=symbols,
            require_symbol=True,
        )
        return Websocket(callback=callback, url=url, **self._ws_kwargs)

    def futures_symbol_mini_ticker(
        self,
        callback: CallbackType,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        """Создает вебсокет для мини‑статистики тикера за последние 24 часа.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            symbol (`str | None`): Один символ для подписки.
            symbols (`Sequence[str] | None`): Список символов для мультиплекс‑подключения.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        url = self._generate_stream_url(
            type="miniTicker",
            url=self._BASE_FUTURES_URL,
            symbol=symbol,
            symbols=symbols,
            require_symbol=True,
        )
        return Websocket(callback=callback, url=url, **self._ws_kwargs)

    def futures_mini_ticker(self, callback: CallbackType) -> Websocket:
        """Создает вебсокет для получения мини-статистики всех тикеров за последние 24 ч.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        url = self._generate_stream_url(type="!miniTicker@arr", url=self._BASE_FUTURES_URL)
        return Websocket(callback=callback, url=url, **self._ws_kwargs)

    def futures_symbol_ticker(
        self,
        callback: CallbackType,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        """Создает вебсокет для расширенной статистики тикера за последние 24 часа.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            symbol (`str | None`): Один символ для подписки.
            symbols (`Sequence[str] | None`): Список символов для мультиплекс‑подключения.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        url = self._generate_stream_url(
            type="ticker",
            url=self._BASE_FUTURES_URL,
            symbol=symbol,
            symbols=symbols,
            require_symbol=True,
        )
        return Websocket(callback=callback, url=url, **self._ws_kwargs)

    def futures_ticker(self, callback: CallbackType) -> Websocket:
        """Создает вебсокет для получения расширенной статистики всех тикеров за последние 24 ч.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        url = self._generate_stream_url(type="!ticker@arr", url=self._BASE_FUTURES_URL)
        return Websocket(callback=callback, url=url, **self._ws_kwargs)

    def futures_book_ticker(
        self,
        callback: CallbackType,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        """Создает вебсокет для получения лучших бид/аск по символам.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            symbol (`str | None`): Один символ для подписки.
            symbols (`Sequence[str] | None`): Список символов для мультиплекс‑подключения.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        url = self._generate_stream_url(
            type="bookTicker",
            url=self._BASE_FUTURES_URL,
            symbol=symbol,
            symbols=symbols,
            require_symbol=True,
        )
        return Websocket(callback=callback, url=url, **self._ws_kwargs)

    def futures_book_depth(
        self,
        callback: CallbackType,
        symbol: str | None,
        levels: BookDepthLevels,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        """Создает вебсокет для получения стакана глубиной N уровней.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            symbol (`str | None`): Один символ для подписки.
            levels (`BookDepthLevels`): Глубина стакана (уровни).
            symbols (`Sequence[str] | None`): Список символов для мультиплекс‑подключения.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        url = self._generate_stream_url(
            type=f"depth{levels}",
            url=self._BASE_FUTURES_URL,
            symbol=symbol,
            symbols=symbols,
            require_symbol=True,
        )
        return Websocket(callback=callback, url=url, **self._ws_kwargs)

    def futures_depth_stream(
        self,
        callback: CallbackType,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        """Создает вебсокет для получения событий изменения стакана (без лимита глубины).

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            symbol (`str | None`): Один символ для подписки.
            symbols (`Sequence[str] | None`): Список символов для мультиплекс‑подключения.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        url = self._generate_stream_url(
            type="depth",
            url=self._BASE_FUTURES_URL,
            symbol=symbol,
            symbols=symbols,
            require_symbol=True,
        )
        return Websocket(callback=callback, url=url, **self._ws_kwargs)

    def futures_mark_price(
        self, callback: CallbackType, interval: MarkPriceUpdateSpeed = "1s"
    ) -> Websocket:
        """Создает вебсокет для получения mark price и funding rate для всех тикеров.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            interval (`MarkPriceUpdateSpeed`): Частота обновления ("1s" или пусто).

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        if interval == "1s":
            type = f"!markPrice@arr@{interval}"
        else:
            type = "!markPrice@arr"
        url = self._generate_stream_url(type=type, url=self._BASE_FUTURES_URL)
        return Websocket(callback=callback, url=url, **self._ws_kwargs)

    def futures_symbol_mark_price(
        self,
        callback: CallbackType,
        interval: MarkPriceUpdateSpeed = "1s",
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        """Создает вебсокет для получения mark price и funding rate по символам.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            interval (`MarkPriceUpdateSpeed`): Частота обновления ("1s" или пусто).
            symbol (`str | None`): Один символ для подписки.
            symbols (`Sequence[str] | None`): Список символов для мультиплекс‑подключения.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        if interval == "1s":
            type = f"markPrice@{interval}"
        else:
            type = "markPrice"
        url = self._generate_stream_url(
            type=type,
            url=self._BASE_FUTURES_URL,
            symbol=symbol,
            symbols=symbols,
            require_symbol=True,
        )
        return Websocket(callback=callback, url=url, **self._ws_kwargs)

    def futures_continuous_klines(
        self,
        callback: CallbackType,
        pair: str,
        contract_type: ContinuousContractType,
        interval: FuturesTimeframe,
    ) -> Websocket:
        """Создает вебсокет для получения свечей по непрерывным контрактам (continuous contract)."""
        url = self._generate_stream_url(
            type=f"{pair.lower()}_{contract_type}@continuousKline_{interval}",
            url=self._BASE_FUTURES_URL,
        )
        return Websocket(callback=callback, url=url, **self._ws_kwargs)

    def liquidation_order(
        self,
        callback: CallbackType,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        """Создает вебсокет для получения ликвидационных ордеров по символам.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            symbol (`str | None`): Один символ для подписки.
            symbols (`Sequence[str] | None`): Список символов для мультиплекс‑подключения.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        url = self._generate_stream_url(
            type="forceOrder",
            url=self._BASE_FUTURES_URL,
            symbol=symbol,
            symbols=symbols,
            require_symbol=True,
        )
        return Websocket(callback=callback, url=url, **self._ws_kwargs)

    def all_liquidation_orders(self, callback: CallbackType) -> Websocket:
        """Создает вебсокет для получения всех ликвидационных ордеров по рынку."""
        url = self._generate_stream_url(type="!forceOrder@arr", url=self._BASE_FUTURES_URL)
        return Websocket(callback=callback, url=url, **self._ws_kwargs)

    def futures_composite_index(
        self,
        callback: CallbackType,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        """Создает вебсокет для получения информации по композитному индексу.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            symbol (`str | None`): Один символ для подписки.
            symbols (`Sequence[str] | None`): Список символов для мультиплекс‑подключения.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        url = self._generate_stream_url(
            type="compositeIndex",
            url=self._BASE_FUTURES_URL,
            symbol=symbol,
            symbols=symbols,
            require_symbol=True,
        )
        return Websocket(callback=callback, url=url, **self._ws_kwargs)

    def futures_contract_info(self, callback: CallbackType) -> Websocket:
        """Создает вебсокет для получения информации о контрактах (Contract Info Stream)."""
        url = self._generate_stream_url(type="!contractInfo", url=self._BASE_FUTURES_URL)
        return Websocket(callback=callback, url=url, **self._ws_kwargs)

    def futures_multi_assets_index(self, callback: CallbackType) -> Websocket:
        """Создает вебсокет для получения индекса активов в режиме Multi-Assets Mode.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        url = self._generate_stream_url(type="!assetIndex@arr", url=self._BASE_FUTURES_URL)
        return Websocket(callback=callback, url=url, **self._ws_kwargs)

    def futures_user_data_stream(self, callback: CallbackType) -> UserWebsocket:
        """Создает вебсокет для получения информации о пользовательских данных.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.

        Возвращает:
            `UserWebsocket`: Вебсокет для получения информации о пользовательских данных.
        """
        if not self.client or not self.client.is_authorized():
            raise NotAuthorized("You must provide authorized client.")
        return UserWebsocket(
            callback=callback, client=self.client, type="FUTURES", **self._ws_kwargs
        )

    def futures_multiplex_socket(self, callback: CallbackType, streams: str) -> Websocket:
        """Создает вебсокет для мультиплексирования нескольких стримов в один.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            streams (`str`): Строка с перечислением стримов.

        Возвращает:
            `Websocket`: Вебсокет для получения информации о пользовательских данных.
        """
        return Websocket(
            callback=callback, url=self._BASE_FUTURES_URL + "?" + streams, **self._ws_kwargs
        )



"""Модуль,который описывает исключения и ошибки, которые могут возникнуть при работе с библиотекой."""

from dataclasses import dataclass


@dataclass
class UniCexException(Exception):
    """Базовое исключение библиотеки."""

    message: str
    """Сообщение об ошибке."""


@dataclass
class NotAuthorized(UniCexException):
    """Исключение, возникающее при отсутствии авторизации."""

    pass


@dataclass
class NotSupported(UniCexException):
    """Исключение, возникающее при попытке использования не поддерживаемой функции."""

    pass


@dataclass
class AdapterError(UniCexException):
    """Исключение, возникающее при ошибке адаптации данных."""

    pass


@dataclass
class QueueOverflowError(UniCexException):
    """Исключение, возникающее при переполнении очереди сообщений."""

    pass



"""`unicex` - библиотека для работы с криптовалютными биржами, реализующая унифицированный интерфейс для работы с различными криптовалютными биржами."""

__all__ = [
    # Mappers
    "get_uni_client",
    "get_uni_websocket_manager",
    # Enums
    "MarketType",
    "Exchange",
    "Timeframe",
    "Side",
    # Types
    "KlineDict",
    "AggTradeDict",
    "TradeDict",
    "TickerDailyDict",
    "RequestMethod",
    "LoggerLike",
    # Interfaces
    "IUniClient",
    "IUniWebsocketManager",
    "IAdapter",
    "IUniAsyncClient",
    "IUniAsyncWebsocketManager",
    # Base clients and websockets
    "Websocket",
    "BaseClient",
    "AsyncWebsocket",
    "BaseAsyncClient",
]

from ._abc import IAdapter, IUniClient, IUniWebsocketManager
from ._abc.asyncio import IUniClient as IUniAsyncClient
from ._abc.asyncio import IUniWebsocketManager as IUniAsyncWebsocketManager
from ._base import BaseClient, Websocket
from ._base.asyncio import BaseClient as BaseAsyncClient
from ._base.asyncio import Websocket as AsyncWebsocket
from .enums import Exchange, MarketType, Side, Timeframe
from .mapper import get_uni_client, get_uni_websocket_manager
from .types import AggTradeDict, KlineDict, LoggerLike, RequestMethod, TickerDailyDict, TradeDict



__all__ = ["ClientMixin"]

import json
import time
from typing import Any

from unicex.types import RequestMethod
from unicex.utils import (
    dict_to_query_string,
    filter_params,
    generate_hmac_sha256_signature,
    sort_params_by_alphabetical_order,
)


class ClientMixin:
    """Миксин для клиентов Bitget API. Содержит общий функционал для работы с REST API Bitget."""

    _BASE_URL: str = "https://api.bitget.com"
    """Базовый URL для REST API Bitget."""

    def _sign_message(
        self,
        method: RequestMethod,
        endpoint: str,
        params: dict[str, Any] | None,
        body: dict[str, Any] | None,
    ) -> tuple[str, str]:
        """Создает timestamp и signature для приватного запроса.

        Алгоритм:
            - формирует строку prehash из timestamp, метода, endpoint, query и body
            - подписывает строку секретным ключом (HMAC-SHA256)
            - кодирует результат в base64

        Параметры:
            method (`RequestMethod`): HTTP-метод (GET, POST и т.д.).
            endpoint (`str`): Относительный путь эндпоинта (например `/api/spot/v1/account/assets`).
            params (`dict[str, Any] | None`): Query-параметры.
            body (`dict[str, Any] | None`): Тело запроса (для POST/PUT).

        Возвращает:
            tuple:
                - `timestamp (str)`: Временная метка в миллисекундах.
                - `signature (str)`: Подпись в формате base64.
        """
        timestamp = str(int(time.time() * 1000))

        path = f"{endpoint}?{dict_to_query_string(params)}" if params else endpoint
        body_str = json.dumps(body) if body else ""
        prehash = f"{timestamp}{method}{path}{body_str}"
        signature = generate_hmac_sha256_signature(
            self._api_secret,  # type: ignore[attr-defined]
            prehash,
            "base64",
        )
        return timestamp, signature

    def _get_headers(self, timestamp: str, signature: str) -> dict[str, str]:
        """Возвращает заголовки для REST-запросов Bitget.

        Параметры:
            timestamp (`str`): Временная метка.
            signature (`str`): Подпись (base64).

        Возвращает:
            `dict[str, str]`: Словарь заголовков запроса.
        """
        headers = {"Content-Type": "application/json", "Accept": "application/json"}
        if self._api_key:  # type: ignore[attr-defined]
            headers.update(
                {
                    "ACCESS-KEY": self._api_key,  # type: ignore[attr-defined]
                    "ACCESS-PASSPHRASE": self._api_passphrase,  # type: ignore[attr-defined]
                    "ACCESS-TIMESTAMP": timestamp,
                    "ACCESS-SIGN": signature,
                }
            )
        return headers

    def _prepare_request_params(
        self,
        *,
        method: RequestMethod,
        endpoint: str,
        signed: bool,
        params: dict[str, Any] | None,
        body: dict[str, Any] | None = None,
    ) -> tuple[str, dict[str, Any] | None, dict[str, Any] | None, dict[str, str] | None]:
        """Готовит данные для запроса.

        Если signed=True:
            - генерирует timestamp и signature
            - добавляет авторизационные заголовки

        Если signed=False:
            - возвращает только url и переданные параметры.

        Параметры:
            method (`RequestMethod`): HTTP-метод (GET, POST и т.д.).
            endpoint (`str`): Относительный путь эндпоинта.
            signed (`bool`): Нужно ли подписывать запрос.
            params (`dict[str, Any] | None`): Query-параметры.
            body (`dict[str, Any] | None`): Тело запроса.

        Возвращает:
            tuple:
                - `url (str)`: Полный URL для запроса.
                - `params (dict | None)`: Query-параметры.
                - `body (dict | None)`: Тело запроса.
                - `headers (dict | None)`: Заголовки (если signed=True).
        """
        url = f"{self._BASE_URL}{endpoint}"

        # Предобрабатывает параметры запроса и сортирует их в соответствии с требованиями Bitget
        if params:
            params = filter_params(params)
            params = sort_params_by_alphabetical_order(params)

        headers = None
        if signed:
            timestamp, signature = self._sign_message(method, endpoint, params, body)
            headers = self._get_headers(timestamp, signature)
        return url, params, body, headers



__all__ = ["UserWebsocketMixin"]


from unicex.exceptions import NotSupported

from ..types import AccountType


class UserWebsocketMixin:
    """Миксин для пользовательского вебсокета Binance. Содержит общий функционал для работы с пользовательским WS API Binance."""

    _BASE_SPOT_URL: str = "wss://stream.binance.com:9443"
    """Базовый URL для вебсокета на спот."""

    _BASE_FUTURES_URL: str = "wss://fstream.binance.com"
    """Базовый URL для вебсокета на фьючерсы."""

    _RENEW_INTERVAL: int = 30 * 60
    """Интервал продления listenKey (сек.)"""

    @classmethod
    def _create_ws_url(cls, type: AccountType, listen_key: str) -> str:
        """Создает URL для подключения к WebSocket."""
        if type == "FUTURES":
            return f"{cls._BASE_FUTURES_URL}/ws/{listen_key}"
        if type == "SPOT":
            return f"{cls._BASE_SPOT_URL}/ws/{listen_key}"
        raise NotSupported(f"Account type '{type}' not supported")



__all__ = ["WebsocketManagerMixin"]

import json
from collections.abc import Sequence
from typing import Literal


class WebsocketManagerMixin:
    """Миксин для менеджеров вебсокетов Bitget."""

    _BASE_URL: str = "wss://ws.bitget.com/v2/ws/public"
    """Базовый URL для вебсокета."""

    def _generate_subscription_message(
        self,
        topic: str,
        market_type: Literal["SPOT", "USDT-FUTURES"],
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> str:
        """Сформировать сообщение для подписки на вебсокет.

        Параметры:
            topic (`str`): Канал подписки (например: "ticker", "candle1m", "depth").
            market_type (`"SPOT" | "USDT-FUTURES"`): Тип рынка для подписки.
            symbol (`str | None`): Торговая пара. Нельзя использовать одновременно с `symbols`.
            symbols (`Sequence[str] | None`): Список торговых пар. Нельзя использовать одновременно с `symbol`.

        Возвращает:
            `str`: JSON-строка с сообщением для подписки на вебсокет.
        """
        if symbol and symbols:
            raise ValueError("Parameters symbol and symbols cannot be used together")
        if not (symbol or symbols):
            raise ValueError("Either symbol or symbols must be provided")

        tickers = [symbol] if symbol else symbols
        streams: list[dict] = [
            {
                "instType": market_type,
                "channel": topic,
                "instId": ticker.upper(),
            }
            for ticker in tickers  # type: ignore
        ]

        return json.dumps({"op": "subscribe", "args": streams})



__all__ = ["WebsocketManagerMixin"]


from collections.abc import Sequence


class WebsocketManagerMixin:
    """Миксин для менеджеров вебсокетов Binance."""

    _BASE_SPOT_URL: str = "wss://stream.binance.com:9443"
    """Базовый URL для вебсокета на спот."""

    _BASE_FUTURES_URL: str = "wss://fstream.binance.com"
    """Базовый URL для вебсокета на фьючерсы."""

    def _generate_stream_url(
        self,
        type: str,
        url: str,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
        require_symbol: bool = False,
    ) -> str:
        """Генерирует URL для вебсокета Binance. Параметры symbol и symbols не могут быть использованы вместе.

        Параметры:
            type (`str`): Тип вебсокета.
            url (`str`): Базовый URL для вебсокета.
            symbol (`str | None`): Символ для подписки.
            symbols (`Sequence[str] | None`): Список символов для подписки.
            require_symbol (`bool`): Требуется ли символ для подписки.

        Возвращает:
            str: URL для вебсокета.
        """
        if symbol and symbols:
            raise ValueError("Parameters symbol and symbols cannot be used together")
        if require_symbol and not (symbol or symbols):
            raise ValueError("Either symbol or symbols must be provided")
        if symbol:
            return f"{url}/ws/{symbol.lower()}@{type}"
        if symbols:
            streams = "/".join(f"{s.lower()}@{type}" for s in symbols)
            return f"{url}/stream?streams={streams}"
        return f"{url}/ws/{type}"



from typing import Literal

BusinessType = Literal["mix", "spot"]
"""Тип торговой линии (контракты / спот)."""

MergeDepthPrecision = Literal["scale0", "scale1", "scale2", "scale3", "scale4", "scale5"]
"""Точность стакана (объединение по цене)."""

MergeDepthLimit = Literal["1", "5", "15", "50", "max"]
"""Лимит объединения стакана."""

OrderbookDepthType = Literal["step0", "step1", "step2", "step3", "step4", "step5"]
"""Тип стакана (объединение по цене)."""

SpotGranularity = Literal[
    "1min",
    "3min",
    "5min",
    "15min",
    "30min",
    "1h",
    "4h",
    "6h",
    "12h",
    "1day",
    "3day",
    "1week",
    "1M",
    "6Hutc",
    "12Hutc",
    "1Dutc",
    "3Dutc",
    "1Wutc",
    "1Mutc",
]
"""Таймфрейм (granularity) для свечей."""

KlineType = Literal["market", "index"]
"""Тип свечей."""

ProductType = Literal["USDT-FUTURES", "COIN-FUTURES", "USDC-FUTURES"]
"""Тип рынков фьючерсов."""

Side = Literal["buy", "sell"]
"""Сторона сделки."""

OrderType = Literal["limit", "market"]
"""Тип ордера."""

TimeInForce = Literal["gtc", "ioc", "fok", "post_only"]
"""Политика исполнения ордера."""

TpslType = Literal["normal", "tpsl"]
"""Тип ордера тейк-профит/стоп-лосс."""

StpMode = Literal["none", "cancel_taker", "cancel_maker", "cancel_both"]
"""Режим стоп-тейк (Self Trade Prevention)."""

BatchMode = Literal["single", "multiple"]
"""Режим батч ордеров."""

PlanType = Literal["amount", "total"]
"""Тип отложенного ордера."""

TriggerType = Literal["fill_price", "mark_price"]
"""Триггер для срабатывания отложенного ордера."""

AssetType = Literal["hold_only", "all"]
"""Тип активов."""

MarginMode = Literal["crossed", "isolated"]
"""Режим маржи."""



"""Пакет содержит синхронные клиенты и менеджеры для работы с биржей Binance."""

__all__ = ["Client", "UniClient", "UserWebsocket", "WebsocketManager", "UniWebsocketManager"]

from .client import Client
from .uni_client import UniClient
from .uni_websocket_manager import UniWebsocketManager
from .user_websocket import UserWebsocket
from .websocket_manager import WebsocketManager



from abc import ABC, abstractmethod
from typing import Any

from unicex.types import AggTradeDict, KlineDict, TickerDailyDict, TradeDict


class IAdapter(ABC):
    """Интерфейс для создания адаптеров, которые преобразовывают ответы с бирж в унифицированный формат."""

    @staticmethod
    @abstractmethod
    def tickers(raw_data: Any, only_usdt: bool) -> list[str]:
        """Преобразует сырой ответ, в котором содержатся данные о тикерах в список тикеров.

        Параметры:
            raw_data (Any): Сырой ответ с биржи.
            only_usdt (bool): Флаг, указывающий, нужно ли включать только тикеры в паре к USDT.

        Возвращает:
            list[str]: Список тикеров.
        """
        pass

    @staticmethod
    @abstractmethod
    def futures_tickers(raw_data: Any, only_usdt: bool) -> list[str]:
        """Преобразует сырой ответ, в котором содержатся данные о тикерах в список тикеров.

        Параметры:
            raw_data (Any): Сырой ответ с биржи.
            only_usdt (bool): Флаг, указывающий, нужно ли включать только тикеры в паре к USDT.

        Возвращает:
            list[str]: Список тикеров.
        """
        pass

    @staticmethod
    @abstractmethod
    def ticker_24h(raw_data: Any) -> dict[str, TickerDailyDict]:
        """Преобразует сырой ответ, в котором содержатся данные о тикере за последние 24 часа
        в унифицированный формат.

        Параметры:
            raw_data (Any): Сырой ответ с биржи.

        Возвращает:
            dict[str, TickerDailyDict]: Словарь, где ключ - тикер, а значение - статистика за последние 24 часа.
        """
        pass

    @staticmethod
    @abstractmethod
    def futures_ticker_24h(raw_data: Any) -> dict[str, TickerDailyDict]:
        """Преобразует сырой ответ, в котором содержатся данные о тикере за последние 24 часа
        в унифицированный формат.

        Параметры:
            raw_data (Any): Сырой ответ с биржи.

        Возвращает:
            dict[str, TickerDailyDict]: Словарь, где ключ - тикер, а значение - статистика за последние 24 часа.
        """
        pass

    @staticmethod
    @abstractmethod
    def last_price(raw_data: Any) -> dict[str, float]:
        """Преобразует сырой ответ, в котором содержатся данные о последних ценах тикеров
        в унифицированный формат.

        Параметры:
            raw_data (Any): Сырой ответ с биржи.

        Возвращает:
            dict[str, float]: Словарь, где ключ - тикер, а значение - последняя цена.
        """
        pass

    @staticmethod
    @abstractmethod
    def futures_last_price(raw_data: Any) -> dict[str, float]:
        """Преобразует сырой ответ, в котором содержатся данные о последних ценах тикеров
        в унифицированный формат.

        Параметры:
            raw_data (Any): Сырой ответ с биржи.

        Возвращает:
            dict[str, float]: Словарь, где ключ - тикер, а значение - последняя цена.
        """
        pass

    @staticmethod
    @abstractmethod
    def klines(raw_data: Any) -> list[KlineDict]:
        """Преобразует сырой ответ, в котором содержатся данные о котировках тикеров в
        унифицированный формат.

        Параметры:
            raw_data (Any): Сырой ответ с биржи.

        Возвращает:
            list[KlineDict]: Список словарей, где каждый словарь содержит данные о свече.
        """
        pass

    @staticmethod
    @abstractmethod
    def futures_klines(raw_data: Any) -> list[KlineDict]:
        """Преобразует сырой ответ, в котором содержатся данные о котировках тикеров в
        унифицированный формат.

        Параметры:
            raw_data (Any): Сырой ответ с биржи.

        Возвращает:
            list[KlineDict]: Список словарей, где каждый словарь содержит данные о свече.
        """
        pass

    @staticmethod
    @abstractmethod
    def funding_rate(raw_data: Any) -> dict[str, float]:
        """Преобразует сырой ответ, в котором содержатся данные о ставках финансирования
        тикеров в унифицированный формат.

        Параметры:
            raw_data (Any): Сырой ответ с биржи.

        Возвращает:
            dict[str, float]: Словарь, где ключ - тикер, а значение - ставка финансирования.
        """
        pass

    @staticmethod
    @abstractmethod
    def klines_message(raw_msg: Any) -> list[KlineDict]:
        """Преобразует сырое сообщение с вебсокета, в котором содержится информация о
        свече/свечах в унифицированный вид.

        Параметры:
            raw_msg (Any): Сырое сообщение с вебсокета.

        Возвращает:
            list[KlineDict]: Список словарей, где каждый словарь содержит данные о свече.
        """
        pass

    @staticmethod
    @abstractmethod
    def futures_klines_message(raw_msg: Any) -> list[KlineDict]:
        """Преобразует сырое сообщение с вебсокета, в котором содержится информация о
        свече/свечах в унифицированный вид.

        Параметры:
            raw_msg (Any): Сырое сообщение с вебсокета.

        Возвращает:
            list[KlineDict]: Список словарей, где каждый словарь содержит данные о свече.
        """
        pass

    @staticmethod
    @abstractmethod
    def aggtrades_message(raw_msg: Any) -> list[AggTradeDict]:
        """Преобразует сырое сообщение вебсокета с агрегированными сделками в унифицированный формат.

        Параметры:
            raw_msg (Any): Сырое сообщение с вебсокета.

        Возвращает:
            list[AggTradeDict]: Список сделок в унифицированном формате.
        """
        pass

    @staticmethod
    @abstractmethod
    def futures_aggtrades_message(raw_msg: Any) -> list[AggTradeDict]:
        """Преобразует сырое сообщение вебсокета с агрегированными сделками в унифицированный формат.

        Параметры:
            raw_msg (Any): Сырое сообщение с вебсокета.

        Возвращает:
            list[AggTradeDict]: Список сделок в унифицированном формате.
        """
        pass

    @staticmethod
    @abstractmethod
    def trades_message(raw_msg: Any) -> list[TradeDict]:
        """Преобразует сырое сообщение вебсокета со сделками в унифицированный формат.

        Параметры:
            raw_msg (Any): Сырое сообщение с вебсокета.

        Возвращает:
            list[TradeDict]: Список сделок в унифицированном формате.
        """
        pass

    @staticmethod
    @abstractmethod
    def futures_trades_message(raw_msg: Any) -> list[TradeDict]:
        """Преобразует сырое сообщение вебсокета со сделками в унифицированный формат.

        Параметры:
            raw_msg (Any): Сырое сообщение с вебсокета.

        Возвращает:
            list[TradeDict]: Список сделок в унифицированном формате.
        """
        pass

    @staticmethod
    @abstractmethod
    def open_interest(raw_data: Any) -> float | dict[str, float]:
        """Преобразует сырое сообщение с вебсокета, в котором содержится информация о
        объеме открытых позиций в унифицированный вид.

        Параметры:
            raw_data (Any): Сырое сообщение с вебсокета.

        Возвращает:
            float | dict[str, float]: Объем открытых позиций в монетах или словарь,
            где ключи - название тикера, а значения - объемы открытых позиций в монетах.
        """
        pass



import json
from collections.abc import Callable, Sequence

from unicex._abc import IWebsocketManager
from unicex.enums import MarketType, Timeframe

from .websocket import BybitWebsocket


class BybitWebsocketManager(IWebsocketManager):
    """Менеджер вебсокетов для Bybit."""

    _BASE_PUBLIC_URL: str = "wss://ws.bitget.com/v2/ws/public"
    _BASE_PRIVATE_URL: str = "wss://stream.bybit.com/v5/private"

    _DEFAULT_RUN_KWARGS = {"ping_payload": json.dumps({"req_id": "100001", "op": "ping"})}

    def klines(
        self,
        market_type: MarketType,
        callback: Callable,
        tickers: Sequence[str],
        timeframe: Timeframe,
    ) -> BybitWebsocket:
        """Возвращает вебсокет для получения свечей."""
        ws = BybitWebsocket(url=self._BASE_PUBLIC_URL, run_kwargs=self._DEFAULT_RUN_KWARGS)
        return ws



from typing import Literal

type Timeframe = Literal[
    "1",
    "3",
    "5",
    "15",
    "30",
    "60",
    "120",
    "240",
    "360",
    "720",
    "D",
    "W",
    "M",
]
"""Возможные интервалы для запросов исторических данных."""

type Category = Literal["linear", "inverse", "spot", "option"]
"""Возможные категории продуктов."""

type FuturesCategory = Literal["linear", "inverse"]
"""Возможные категории фьючерсных продуктов."""

type DerivativesCategory = Literal["linear", "inverse", "option"]
"""Возможные категории деривативов."""

type Side = Literal["Buy", "Sell"]
"""Возможные стороны для торговли."""

type OrderType = Literal[
    "Market",
    "Limit",
    "Conditional",
]
"""Возможные типы ордеров."""

type TimeInForce = Literal["GTC", "IOC", "FOK", "PostOnly", "RPI"]
"""Возможные стратегии времени действия ордера."""

type OrderStatus = Literal[
    "New",
    "PartiallyFilled",
    "Filled",
    "Cancelled",
    "Rejected",
    "PartiallyFilledCanceled",
    "Deactivated",
    "Triggered",
    "Untriggered",
    "Active",
]
"""Возможные статусы ордеров."""

type OrderFilter = Literal["Order", "tpslOrder", "StopOrder"]
"""Фильтры типов ордеров для спота."""

type TriggerBy = Literal["LastPrice", "IndexPrice", "MarkPrice"]
"""Типы цен для триггера условных ордеров."""

type PositionIdx = Literal[0, 1, 2]
"""Индексы позиций: 0 - одностороння, 1 - хедж Buy, 2 - хедж Sell."""

# ========== SLIPPAGE ==========

type SlippageToleranceType = Literal["TickSize", "Percent"]
"""Типы толерантности к проскальзыванию."""

# ========== TAKE PROFIT / STOP LOSS ==========

type TpSlMode = Literal["Full", "Partial"]
"""Режимы работы Take Profit / Stop Loss."""

type TpSlOrderType = Literal["Market", "Limit"]
"""Типы ордеров для TP/SL."""

# ========== POSITIONS ==========

type PositionSide = Literal["Buy", "Sell", "None"]
"""Стороны позиций."""

type MarginMode = Literal["ISOLATED_MARGIN", "REGULAR_MARGIN", "PORTFOLIO_MARGIN"]
"""Режимы маржи."""

# ========== ACCOUNT ==========

type AccountType = Literal["UNIFIED", "NORMAL", "FUND"]
"""Типы аккаунтов."""

type UnifiedMarginStatus = Literal["1", "2", "3", "4"]
"""Статусы объединенной маржи."""

# ========== TRANSFERS ==========

type TransferStatus = Literal["SUCCESS", "PENDING", "FAILED"]
"""Статусы переводов."""

type AccountType_Transfer = Literal["FUND", "UNIFIED", "SPOT", "CONTRACT", "OPTION"]
"""Типы аккаунтов для переводов."""

# ========== MARKET UNIT ==========

type MarketUnit = Literal["baseCoin", "quoteCoin"]
"""Единицы измерения для рыночных ордеров на споте."""

# ========== SMP (Self-Match Prevention) ==========

type SmpType = Literal["None", "CancelMaker", "CancelTaker", "CancelBoth"]
"""Типы предотвращения самопересечения."""

# ========== WALLET ==========

type CoinType = Literal["FUND", "UNIFIED"]
"""Типы кошельков."""

# ========== LEVERAGE ==========

type LeverageTokenLimitType = Literal["PurchaseLimit", "RedemptionLimit"]
"""Типы лимитов для токенов с кредитным плечом."""

# ========== INSURANCE ==========

type InsuranceCoin = Literal["BTC", "USDT"]
"""Монеты страхового фонда."""

# ========== HISTORICAL DATA ==========

type DataRecordingPeriod = Literal["5min", "15min", "30min", "1h", "4h", "1d"]
"""Периоды записи исторических данных."""

# ========== ANNOUNCEMENT ==========

type AnnouncementType = Literal["new_crypto", "delist", "latest"]
"""Типы объявлений."""

type AnnouncementTag = Literal[
    "Spot",
    "Derivatives",
    "Options",
    "Leveraged_Tokens",
    "Launchpad",
    "Launchpool",
    "Spot_Listings",
    "VIP",
    "Margin_Trading",
    "P2P",
    "Trading_Bots",
]
"""Теги объявлений."""

# ========== RISK LIMIT ==========

type RiskLimitCategory = Literal["linear", "inverse"]
"""Категории для риск лимитов."""

# ========== DELIVERY PRICE ==========

type DeliveryPriceCategory = Literal["linear", "inverse", "option"]
"""Категории для цен поставки."""

# ========== FUNDING ==========

type FundingIntervalCategory = Literal["linear", "inverse"]
"""Категории для интервалов фондирования."""

# ========== OPTIONS ==========

type OptionType = Literal["Call", "Put"]
"""Типы опционов."""

# ========== EXECUTION TYPES ==========

type ExecType = Literal[
    "Trade", "AdlTrade", "Funding", "BustTrade", "Settle", "Delivery", "BlockTrade", "MovePosition"
]
"""Типы исполнения сделок."""

# ========== INTERVALS FOR STATISTICS ==========

type StatsInterval = Literal["5min", "15min", "30min", "1h", "4h", "1d"]
"""Интервалы для статистики."""

# ========== API KEY PERMISSIONS ==========

type Permission = Literal[
    "ContractTrade",
    "Spot",
    "Wallet",
    "Options",
    "Derivatives",
    "CopyTrading",
    "BlockTrade",
    "Exchange",
    "NFT",
]
"""Разрешения для API ключей."""

# ========== UNIFIED UPGRADE STATUS ==========

type UpgradeStatus = Literal["0", "1", "2", "3", "4", "5"]
"""Статусы обновления до единого аккаунта."""

# ========== BORROWING ==========

type InterestBearingBorrowSize = Literal["0", "1"]
"""Размер заимствования под проценты."""

# ========== INSTITUTIONAL LENDING ==========

type LendingProductType = Literal["FLEX_TERM", "FIXED_TERM"]
"""Типы продуктов институционального кредитования."""

type LendingOrderStatus = Literal["1", "2", "3", "4", "5", "6"]
"""Статусы ордеров институционального кредитования."""

# ========== C2C LENDING ==========

type LendingOrderType = Literal["1", "2"]
"""Типы ордеров C2C кредитования."""

# ========== CONVERT ==========

type ConvertStatus = Literal["init", "processing", "success", "failure"]
"""Статусы конвертации."""

# ========== PRE-UPGRADE ==========

type UpgradeToUTAStatus = Literal["0", "1", "2", "3"]
"""Статусы предварительного обновления до UTA."""

# ========== RESPONSE FORMATS ==========

type RetCode = Literal[
    0, 10001, 10002, 10003, 10004, 10005, 10006, 10007, 10009, 10010, 10016, 10017, 10018
]
"""Коды возврата API."""

# ========== WEBSOCKET ==========

type WSMessageType = Literal["snapshot", "delta"]
"""Типы сообщений WebSocket."""

type WSOperation = Literal["subscribe", "unsubscribe", "ping", "pong"]
"""Операции WebSocket."""

type WSTopicType = Literal[
    "orderbook", "publicTrade", "ticker", "kline", "liquidation", "ltNav", "ltTicker", "ltKline"
]
"""Типы публичных топиков WebSocket."""

type WSPrivateTopicType = Literal["position", "execution", "order", "wallet", "greeks"]
"""Типы приватных топиков WebSocket."""

# ========== DEPOSIT/WITHDRAW ==========

type DepositStatus = Literal["0", "1", "2", "3", "4"]
"""Статусы депозитов."""

type WithdrawStatus = Literal[
    "SecurityCheck", "Pending", "success", "CancelByUser", "Reject", "Fail", "MoreInformation"
]
"""Статусы выводов средств."""

type WithdrawType = Literal["0", "1", "2"]
"""Типы выводов: 0-on-chain, 1-internal, 2-off-chain."""
